"""Portfolio module.
`Portfolio` is returned by the `predict` method of Optimization estimators.
It needs to be homogenous to the convex optimization problems meaning that `Portfolio`
is the dot product of the assets weights with the assets returns.
"""

# Copyright (c) 2023
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# License: BSD 3 clause

import numbers
from typing import ClassVar

import numpy as np
import numpy.typing as npt
import pandas as pd
import plotly.express as px

import skfolio.typing as skt
from skfolio.measures import RiskMeasure, effective_number_assets
from skfolio.portfolio._base import _ZERO_THRESHOLD, BasePortfolio
from skfolio.utils.tools import (
    args_names,
    cached_property_slots,
    default_asset_names,
    input_to_array,
)

pd.options.plotting.backend = "plotly"


class Portfolio(BasePortfolio):
    r"""
    Portfolio class.

    `Portfolio` is returned by the `predict` method of Optimization estimators.
    It is homogenous to the convex optimization problems meaning that `Portfolio` is
    the dot product of the assets weights with the assets returns.

    Parameters
    ----------
    X : array-like of shape (n_observations, n_assets)
        Price returns of the assets.
        If `X` is a DataFrame or another array containers that implements 'columns'
        and 'index', the columns will be considered as assets names and the
        indices will be considered as observations.
        Otherwise, we use `["x0", "x1", ..., "x(n_assets - 1)"]` as asset names
        and `[0, 1, ..., n_observations]` as observations.

    weights : array-like of shape (n_assets,) | dict[str, float]
        Portfolio weights.
        If a dictionary is provided, its (key/value) pair must be the
        (asset name/asset weight) and `X` must be a DataFrame with assets names
        in columns.

    transaction_costs : float | dict[str, float] | array-like of shape (n_assets, ), optional
        Linear transaction costs of the assets. The Portfolio total transaction cost
        is:

        .. math:: total\_cost = \sum_{i=1}^{N} c_{i} \times |w_{i} - w\_prev_{i}|

        with :math:`c_{i}` the transaction cost of asset i, :math:`w_{i}` its weight
        and :math:`w\_prev_{i}` its previous weight (defined in `previous_weights`).
        The float :math:`total\_cost` is used in the portfolio returns:

        .. math:: ptf\_returns = R \cdot w - total\_cost

        with :math:`R` the matrix af assets returns and :math:`w` the vector of
        assets weights.

        If a float is provided, it is applied to each asset.
        If a dictionary is provided, its (key/value) pair must be the
        (asset name/asset weight) and `X` must be a DataFrame with assets names
        in columns.
        The default (`None`) means no transaction costs.

        .. warning::

            To be homogenous to the optimization problems, the periodicity of the
            transaction costs needs to be homogenous to the periodicity of the
            returns `X`. For example, if `X` is composed of **daily** returns,
            the `transaction_costs` need to be expressed in **daily** transaction costs.

    management_fees : float | dict[str, float] | array-like of shape (n_assets, ), optional
        Linear management fees of the assets. The Portfolio total management cost
        is:

        .. math:: total\_fee = \sum_{i=1}^{N} f_{i} \times w_{i}

        with :math:`f_{i}` the management fee of asset i and :math:`w_{i}` its weight.
        The float :math:`total\_fee` is used in the portfolio returns:

        .. math:: ptf\_returns = R \cdot w - total\_fee

        with :math:`R` the matrix af assets returns and :math:`w` the vector of
        assets weights.

        If a float is provided, it is applied to each asset.
        If a dictionary is provided, its (key/value) pair must be the
        (asset name/asset weight) and `X` must be a DataFrame with assets names
        in columns.
        The default (`None`) means no management fees.

        .. warning::

            To be homogenous to the optimization problems, the periodicity of the
            management fees needs to be homogenous to the periodicity of the
            returns `X`. For example, if `X` is composed of **daily** returns,
            the `management_fees` need to be expressed in **daily** fees.

    previous_weights : float | dict[str, float] | array-like of shape (n_assets, ), optional
        Previous portfolio weights.
        Previous weights are used to compute the total portfolio cost.
        If `transaction_costs` is 0, `previous_weights` will have no impact.
        If a float is provided, it is applied to each asset.
        If a dictionary is provided, its (key/value) pair must be the
        (asset name/asset previous weight) and `X` must be a DataFrame with assets names
        in columns.
        The default (`None`) means no previous weights.

    name : str, optional
        Name of the portfolio.
        The default (`None`) is to use the object id.

    tag : str, optional
        Tag given to the portfolio.
        Tags are used to manipulate groups of Portfolios from a `Population`.

    fitness_measures : list[measures], optional
        List of fitness measures.
        Fitness measures are used to compute the portfolio fitness which is used to
        compute domination.
        The default (`None`) is to use the list [PerfMeasure.MEAN, RiskMeasure.VARIANCE]

    annualized_factor : float, default=252.0
        Factor used to annualize the below measures using the square-root rule:

            * Annualized Mean = Mean * factor
            * Annualized Variance = Variance * factor
            * Annualized Semi-Variance = Semi-Variance * factor
            * Annualized Standard-Deviation = Standard-Deviation * sqrt(factor)
            * Annualized Semi-Deviation = Semi-Deviation * sqrt(factor)
            * Annualized Sharpe Ratio = Sharpe Ratio * sqrt(factor)
            * Annualized Sortino Ratio = Sortino Ratio * sqrt(factor)

    risk_free_rate : float, default=0.0
        Risk-free rate. The default value is `0.0`.

    compounded : bool, default=False
        If this is set to True, cumulative returns are compounded.
        The default is `False`.

    min_acceptable_return : float, optional
        The minimum acceptable return used to distinguish "downside" and "upside"
        returns for the computation of lower partial moments:

            * First Lower Partial Moment
            * Semi-Variance
            * Semi-Deviation

        The default (`None`) is to use the mean.

    value_at_risk_beta : float, default=0.95
        The confidence level of the Portfolio VaR (Value At Risk) which represents
        the return on the worst (1-beta)% observations.
        The default value is `0.95`.

    entropic_risk_measure_theta : float, default=1.0
        The risk aversion level of the Portfolio Entropic Risk Measure.
        The default value is `1.0`.

    entropic_risk_measure_beta : float, default=0.95
        The confidence level of the Portfolio Entropic Risk Measure.
        The default value is `0.95`.

    cvar_beta : float, default=0.95
        The confidence level of the Portfolio CVaR (Conditional Value at Risk) which
        represents the expected VaR on the worst (1-beta)% observations.
        The default value is `0.95`.

    evar_beta : float, default=0.95
        The confidence level of the Portfolio EVaR (Entropic Value at Risk).
        The default value is `0.95`.

    drawdown_at_risk_beta : float, default=0.95
        The confidence level of the Portfolio Drawdown at Risk (DaR) which represents
        the drawdown on the worst (1-beta)% observations.
        The default value is `0.95`.

    cdar_beta : float, default=0.95
        The confidence level of the Portfolio CDaR (Conditional Drawdown at Risk) which
        represents the expected drawdown on the worst (1-beta)% observations.
        The default value is `0.95`.

    edar_beta : float, default=0.95
        The confidence level of the Portfolio EDaR (Entropic Drawdown at Risk).
        The default value is `0.95`.

    Attributes
    ----------
    n_observations : float
        Number of observations.

    mean : float
        Mean of the portfolio returns.

    annualized_mean : float
        Mean annualized by :math:`mean \times annualization\_factor`

    mean_absolute_deviation : float
        Mean Absolute Deviation. The deviation is the difference between the
        return and a minimum acceptable return (`min_acceptable_return`).

    first_lower_partial_moment : float
        First Lower Partial Moment. The First Lower Partial Moment is the mean of the
        returns below a minimum acceptable return (`min_acceptable_return`).

    variance : float
        Variance (Second Moment)

    annualized_variance : float
        Variance annualized by :math:`variance \times annualization\_factor`

    semi_variance : float
        Semi-variance (Second Lower Partial Moment).
        The semi-variance is the variance of the returns below a minimum acceptable
        return (`min_acceptable_return`).

    annualized_semi_variance : float
        Semi-variance annualized by
        :math:`semi\_variance \times annualization\_factor`

    standard_deviation : float
        Standard Deviation (Square Root of the Second Moment).

    annualized_standard_deviation : float
        Standard Deviation annualized by
        :math:`standard\_deviation \times \sqrt{annualization\_factor}`

    semi_deviation : float
        Semi-deviation (Square Root of the Second Lower Partial Moment).
        The Semi Standard Deviation is the Standard Deviation of the returns below a
        minimum acceptable return (`min_acceptable_return`).

    annualized_semi_deviation : float
        Semi-deviation annualized by
        :math:`semi\_deviation \times \sqrt{annualization\_factor}`

    skew : float
        Skew. The Skew is a measure of the lopsidedness of the distribution.
        A symmetric distribution have a Skew of zero.
        Higher Skew corresponds to longer right tail.

    kurtosis : float
        Kurtosis. It is a measure of the heaviness of the tail of the distribution.
        Higher Kurtosis corresponds to greater extremity of deviations (fat tails).

    fourth_central_moment : float
       Fourth Central Moment.

    fourth_lower_partial_moment : float
        Fourth Lower Partial Moment. It is a measure of the heaviness of the downside
        tail of the returns below a minimum acceptable return (`min_acceptable_return`).
        Higher Fourth Lower Partial Moment corresponds to greater extremity of downside
        deviations (downside fat tail).

    worst_realization : float
        Worst Realization which is the worst return.

    value_at_risk : float
        Historical VaR (Value at Risk).
        The VaR is the maximum loss at a given confidence level (`value_at_risk_beta`).

    cvar : float
        Historical CVaR (Conditional Value at Risk). The CVaR (or Tail VaR) represents
        the mean shortfall at a specified confidence level (`cvar_beta`).

    entropic_risk_measure : float
        Historical Entropic Risk Measure. It is a risk measure which depends on the
        risk aversion defined by the investor (`entropic_risk_measure_theta`) through
        the exponential utility function at a given confidence level
        (`entropic_risk_measure_beta`).

    evar : float
         Historical EVaR (Entropic Value at Risk). It is a coherent risk measure which
         is an upper bound for the VaR and the CVaR, obtained from the Chernoff
         inequality at a given confidence level (`evar_beta`). The EVaR can be
         represented by using the concept of relative entropy.

    drawdown_at_risk : float
        Historical Drawdown at Risk. It is the maximum drawdown at a given
        confidence level (`drawdown_at_risk_beta`).

    cdar : float
        Historical CDaR (Conditional Drawdown at Risk) at a given confidence level
        (`cdar_beta`).

    max_drawdown : float
        Maximum Drawdown.

    average_drawdown : float
        Average Drawdown.

    edar : float
        EDaR (Entropic Drawdown at Risk). It is a coherent risk measure which is an
        upper bound for the Drawdown at Risk and the CDaR, obtained from the Chernoff
        inequality at a given confidence level (`edar_beta`). The EDaR can be
        represented by using the concept of relative entropy.

    ulcer_index : float
        Ulcer Index

    gini_mean_difference : float
        Gini Mean Difference (GMD). It is the expected absolute difference between two
        realizations. The GMD is a superior measure of variability  for non-normal
        distribution than the variance. It can be used to form necessary conditions
        for second-degree stochastic dominance, while the variance cannot.

    mean_absolute_deviation_ratio : float
        Mean Absolute Deviation ratio.
        It is the excess mean (mean - risk_free_rate) divided by the MaD.

    first_lower_partial_moment_ratio : float
        First Lower Partial Moment ratio.
        It is the excess mean (mean - risk_free_rate) divided by the First Lower
        Partial Moment.

    sharpe_ratio : float
        Sharpe ratio.
        It is the excess mean (mean - risk_free_rate) divided by the standard-deviation.

    annualized_sharpe_ratio : float
        Sharpe ratio annualized by
        :math:`sharpe\_ratio \times \sqrt{annualization\_factor}`.

    sortino_ratio : float
        Sortino ratio.
        It is the excess mean (mean - risk_free_rate) divided by the semi
        standard-deviation.

    annualized_sortino_ratio : float
        Sortino ratio annualized by
        :math:`sortino\_ratio \times \sqrt{annualization\_factor}`.

    value_at_risk_ratio : float
        VaR ratio.
        It is the excess mean (mean - risk_free_rate) divided by the Value at Risk
        (VaR).

    cvar_ratio : float
        CVaR ratio.
        It is the excess mean (mean - risk_free_rate) divided by the Conditional Value
        at Risk (CVaR).

    entropic_risk_measure_ratio : float
        Entropic risk measure ratio.
        It is the excess mean (mean - risk_free_rate) divided by the Entropic risk
        measure.

    evar_ratio : float
        EVaR ratio.
        It is the excess mean (mean - risk_free_rate) divided by the EVaR (Entropic
        Value at Risk).

    worst_realization_ratio : float
        Worst Realization ratio.
        It is the excess mean (mean - risk_free_rate) divided by the Worst Realization
        (worst return).

    drawdown_at_risk_ratio : float
        Drawdown at Risk ratio.
        It is the excess mean (mean - risk_free_rate) divided by the drawdown at
        risk.

    cdar_ratio : float
        CDaR ratio.
        It is the excess mean (mean - risk_free_rate) divided by the CDaR (conditional
        drawdown at risk).

    calmar_ratio : float
        Calmar ratio.
        It is the excess mean (mean - risk_free_rate) divided by the Maximum Drawdown.

    average_drawdown_ratio : float
        Average Drawdown ratio.
        It is the excess mean (mean - risk_free_rate) divided by the Average Drawdown.

    edar_ratio : float
        EDaR ratio.
        It is the excess mean (mean - risk_free_rate) divided by the EDaR (Entropic
        Drawdown at Risk).

    ulcer_index_ratio : float
        Ulcer Index ratio.
        It is the excess mean (mean - risk_free_rate) divided by the Ulcer Index.

    gini_mean_difference_ratio : float
        Gini Mean Difference ratio.
        It is the excess mean (mean - risk_free_rate) divided by the Gini Mean
        Difference.
    """

    _read_only_attrs: ClassVar[set] = BasePortfolio._read_only_attrs.copy()
    _read_only_attrs.update(
        {
            "X",
            "assets",
            "weights",
            "previous_weights",
            "transaction_costs",
            "management_fees",
            "n_assets",
            "total_cost",
            "total_fee",
        }
    )

    __slots__ = {
        # read-only
        "X",
        "weights",
        "previous_weights",
        "transaction_costs",
        "management_fees",
        "assets",
        "n_assets",
        "total_cost",
        "total_fee",
        # custom getter (read-only and cached)
        "_nonzero_assets",
        "_nonzero_assets_index",
    }

    def __init__(
        self,
        X: npt.ArrayLike,
        weights: skt.MultiInput,
        previous_weights: skt.MultiInput = None,
        transaction_costs: skt.MultiInput = None,
        management_fees: skt.MultiInput = None,
        risk_free_rate: float = 0,
        name: str | None = None,
        tag: str | None = None,
        annualized_factor: float = 252,
        fitness_measures: list[skt.Measure] | None = None,
        compounded: bool = False,
        min_acceptable_return: float | None = None,
        value_at_risk_beta: float = 0.95,
        entropic_risk_measure_theta: float = 1,
        entropic_risk_measure_beta: float = 0.95,
        cvar_beta: float = 0.95,
        evar_beta: float = 0.95,
        drawdown_at_risk_beta: float = 0.95,
        cdar_beta: float = 0.95,
        edar_beta: float = 0.95,
    ):
        # extract assets names from X
        assets = None
        observations = None
        if hasattr(X, "columns"):
            assets = np.asarray(X.columns, dtype=object)
            observations = np.asarray(X.index)

        # We don't perform extensive checks (like in check_X) for faster instantiation.
        rets = np.asarray(X)
        if rets.ndim != 2:
            raise ValueError("`X` must be a 2D array-like")

        n_observations, n_assets = rets.shape

        weights = input_to_array(
            items=weights,
            n_assets=n_assets,
            fill_value=0,
            dim=1,
            assets_names=assets,
            name="weights",
        )

        if previous_weights is None:
            previous_weights = np.zeros(n_assets)
        else:
            previous_weights = input_to_array(
                items=previous_weights,
                n_assets=n_assets,
                fill_value=0,
                dim=1,
                assets_names=assets,
                name="previous_weights",
            )

        if transaction_costs is None:
            transaction_costs = 0
        elif not np.isscalar(transaction_costs):
            transaction_costs = input_to_array(
                items=transaction_costs,
                n_assets=n_assets,
                fill_value=0,
                dim=1,
                assets_names=assets,
                name="transaction_costs",
            )

        if management_fees is None:
            management_fees = 0
        elif not np.isscalar(management_fees):
            management_fees = input_to_array(
                items=management_fees,
                n_assets=n_assets,
                fill_value=0,
                dim=1,
                assets_names=assets,
                name="management_fees",
            )

        # Default observations and assets if X is not a DataFrame
        if observations is None or len(observations) == 0:
            observations = np.arange(n_observations)

        if assets is None or len(assets) == 0:
            assets = default_asset_names(n_assets=n_assets)

        # Computing portfolio returns
        if np.isscalar(transaction_costs) and transaction_costs == 0:
            total_cost = 0
        else:
            total_cost = (transaction_costs * abs(previous_weights - weights)).sum()

        if np.isscalar(management_fees) and management_fees == 0:
            total_fee = 0
        else:
            total_fee = (management_fees * weights).sum()

        returns = weights @ rets.T - total_cost - total_fee

        if np.any(np.isnan(returns)):
            raise ValueError("NaN found in `returns`")

        super().__init__(
            returns=returns,
            observations=observations,
            name=name,
            tag=tag,
            fitness_measures=fitness_measures,
            compounded=compounded,
            risk_free_rate=risk_free_rate,
            annualized_factor=annualized_factor,
            min_acceptable_return=min_acceptable_return,
            value_at_risk_beta=value_at_risk_beta,
            cvar_beta=cvar_beta,
            entropic_risk_measure_theta=entropic_risk_measure_theta,
            entropic_risk_measure_beta=entropic_risk_measure_beta,
            evar_beta=evar_beta,
            drawdown_at_risk_beta=drawdown_at_risk_beta,
            cdar_beta=cdar_beta,
            edar_beta=edar_beta,
        )
        self._loaded = False
        # We save the original array-like object and not the numpy copy for improved
        # memory
        self.X = X
        self.assets = assets
        self.n_assets = n_assets
        self.weights = weights
        self.transaction_costs = transaction_costs
        self.management_fees = management_fees
        self.previous_weights = previous_weights
        self.total_cost = total_cost
        self.total_fee = total_fee
        self._loaded = True

    def __neg__(self):
        args = {arg: getattr(self, arg) for arg in args_names(self.__init__)}
        args["weights"] = -self.weights
        return self.__class__(**args)

    def __abs__(self):
        args = {arg: getattr(self, arg) for arg in args_names(self.__init__)}
        args["weights"] = np.abs(self.weights)
        return self.__class__(**args)

    def __round__(self, n: int):
        args = {arg: getattr(self, arg) for arg in args_names(self.__init__)}
        args["weights"] = np.round(self.weights, n)
        return self.__class__(**args)

    def __add__(self, other):
        if not isinstance(other, Portfolio):
            raise TypeError(
                f"Cannot add a Portfolio with an object of type {type(other)}"
            )
        args = args_names(self.__init__)
        for arg in args:
            if arg not in ["weights", "name", "tag"] and not np.array_equal(
                getattr(self, arg), getattr(other, arg)
            ):
                raise ValueError(f"Cannot add two Portfolios with different `{arg}`")
        args = {arg: getattr(self, arg) for arg in args}
        args["weights"] = self.weights + other.weights
        return self.__class__(**args)

    def __sub__(self, other):
        if not isinstance(other, Portfolio):
            raise TypeError(
                f"Cannot add a Portfolio with an object of type {type(other)}"
            )
        args = args_names(self.__init__)
        for arg in args:
            if arg not in ["weights", "name", "tag"] and not np.array_equal(
                getattr(self, arg), getattr(other, arg)
            ):
                raise ValueError(
                    f"Cannot subtract two Portfolios with different `{arg}`"
                )
        args = {arg: getattr(self, arg) for arg in args}
        args["weights"] = self.weights - other.weights
        return self.__class__(**args)

    def __mul__(self, other: numbers.Number):
        if not isinstance(other, numbers.Number):
            raise TypeError(
                "Portfolio can only be multiplied by a number, but received a"
                f" {type(other)}"
            )
        args = {arg: getattr(self, arg) for arg in args_names(self.__init__)}
        args["weights"] = other * self.weights
        return self.__class__(**args)

    __rmul__ = __mul__

    def __floordiv__(self, other: numbers.Number):
        if not isinstance(other, numbers.Number):
            raise TypeError(
                "Portfolio can only be floor divided by a number, but received a"
                f" {type(other)}"
            )
        args = {arg: getattr(self, arg) for arg in args_names(self.__init__)}
        args["weights"] = np.floor_divide(self.weights, other)
        return self.__class__(**args)

    def __truediv__(self, other: numbers.Number):
        if not isinstance(other, numbers.Number):
            raise TypeError(
                "Portfolio can only be divided by a number, but received a"
                f" {type(other)}"
            )
        args = {arg: getattr(self, arg) for arg in args_names(self.__init__)}
        args["weights"] = self.weights / other
        return self.__class__(**args)

    # Custom attribute getter (read-only and cached)
    @cached_property_slots
    def nonzero_assets(self) -> np.ndarray:
        """Invested asset :math:`abs(weights) > 0.001%`"""
        return self.assets[self.nonzero_assets_index]

    @cached_property_slots
    def nonzero_assets_index(self) -> np.ndarray:
        """Indices of invested asset :math:`abs(weights) > 0.001%`"""
        return np.flatnonzero(abs(self.weights) > _ZERO_THRESHOLD)

    @property
    def composition(self) -> pd.DataFrame:
        """DataFrame of the Portfolio composition."""
        weights = self.weights[self.nonzero_assets_index]
        df = pd.DataFrame({"asset": self.nonzero_assets, "weight": weights})
        df.sort_values(by="weight", ascending=False, inplace=True)
        df.rename(columns={"weight": self.name}, inplace=True)
        df.set_index("asset", inplace=True)
        return df

    @property
    def diversification(self):
        """Weighted average of volatility divided by the portfolio volatility."""
        return (
            self.weights @ np.std(np.asarray(self.X), axis=0) / self.standard_deviation
        )

    @property
    def sric(self) -> float:
        """Sharpe Ratio Information Criterion (SRIC).

        It is an unbiased estimator of the Sharpe Ratio adjusting for both sources of
        bias which are noise fit and estimation error [1]_.

        References
        ----------
        .. [1]  "Noise Fit, Estimation Error and a Sharpe Information Criterion",
            Dirk Paulsen (2019)
        """
        return self.sharpe_ratio - self.n_assets / (
            self.n_observations * self.sharpe_ratio
        )

    @property
    def effective_number_assets(self) -> float:
        r"""Computes the effective number of assets, defined as the inverse of the
        Herfindahl index [1]_:

        .. math:: N_{eff} = \frac{1}{\Vert w \Vert_{2}^{2}}

        It quantifies portfolio concentration, with a higher value indicating a more
        diversified portfolio.

        Returns
        -------
        value : float
            Effective number of assets.

        References
        ----------
        .. [1] "Banking and Financial Institutions Law in a Nutshell".
            Lovett, William Anthony (1988)
        """
        return effective_number_assets(weights=self.weights)

    # Public methods
    def expected_returns_from_assets(
        self, assets_expected_returns: np.ndarray
    ) -> float:
        """Compute the Portfolio expected returns from the assets expected returns,
        weights, management costs and transaction fees.

        Parameters
        ----------
        assets_expected_returns : ndarray of shape (n_assets,)
            The vector of assets expected returns.

        Returns
        -------
        value : float
            The Portfolio expected returns.
        """
        return (
            self.weights @ assets_expected_returns.T - self.total_cost - self.total_fee
        )

    def variance_from_assets(self, assets_covariance: np.ndarray) -> float:
        """Compute the Portfolio variance expectation from the assets covariance and
        weights.

        Parameters
        ----------
        assets_covariance : ndarray of shape (n_assets,n_assets)
            The matrix of assets covariance expectation.

        Returns
        -------
        value : float
            The Portfolio variance from the assets covariance.
        """
        return float(self.weights @ assets_covariance @ self.weights.T)

    def contribution(
        self, measure: skt.Measure, spacing: float | None = None
    ) -> np.ndarray:
        r"""Compute the contribution of each asset to a given measure.

        Parameters
        ----------
        measure : Measure
            The measure used for the contribution computation.

        spacing : float, optional
            Spacing "h" of the finite difference:
            :math:`contribution(wi)= \frac{measure(wi-h) - measure(wi+h)}{2h}`

        Returns
        -------
        values : ndrray of shape (n_assets,)
            The measure contribution of each asset.
        """
        if spacing is None:
            if measure in [
                RiskMeasure.MAX_DRAWDOWN,
                RiskMeasure.AVERAGE_DRAWDOWN,
                RiskMeasure.CDAR,
                RiskMeasure.EDAR,
            ]:
                spacing = 1e-1
            else:
                spacing = 1e-5
        args = {arg: getattr(self, arg) for arg in args_names(self.__init__)}

        def get_risk(i: int, h: float) -> float:
            a = args.copy()
            w = a["weights"].copy()
            w[i] += h
            a["weights"] = w
            return getattr(Portfolio(**a), measure.value)

        cont = [
            (get_risk(i, h=spacing) - get_risk(i, h=-spacing))
            / (2 * spacing)
            * self.weights[i]
            for i in range(len(self.weights))
        ]
        return np.array(cont)

    def plot_contribution(self, measure: skt.Measure, spacing: float | None = None):
        r"""Plot the contribution of each asset to a given measure.

        Parameters
        ----------
        measure : Measure
            The measure used for the contribution computation.

        spacing : float, optional
            Spacing "h" of the finite difference:
            :math:`contribution(wi)= \frac{measure(wi-h) - measure(wi+h)}{2h}`

        Returns
        -------
        plot : Figure
            The plotly Figure of assets contribution to the measure.
        """
        cont = self.contribution(measure=measure, spacing=spacing)
        df = pd.DataFrame(cont, index=self.assets, columns=["contribution"])
        fig = px.bar(df, x=df.index, y=df.columns)
        fig.update_layout(
            title=f"{measure} contribution",
            xaxis_title="Asset",
            yaxis_title=f"{measure} contribution",
        )
        return fig

    def summary(self, formatted: bool = True) -> pd.Series:
        """Portfolio summary of all its measures.

        Parameters
        ----------
        formatted : bool, default=True
            If this is set to True, the measures are formatted into rounded string
            with units.

        Returns
        -------
        summary : series
            Portfolio summary.
        """
        df = super().summary(formatted=formatted)
        assets_number = self.n_assets
        effective_nb_assets = self.effective_number_assets
        if formatted:
            assets_number = str(assets_number)
            effective_nb_assets = str(effective_nb_assets)
        df["Effective Number of Assets"] = effective_nb_assets
        df["Assets Number"] = assets_number
        return df

    def get_weight(self, asset: str) -> float:
        """Get the weight of a given asset.

        Parameters
        ----------
        asset : str
            Name of the asset.

        Returns
        -------
        weight : float
            Weight of the asset.
        """
        try:
            return self.weights[np.where(self.assets == asset)[0][0]]
        except IndexError:
            raise IndexError("{asset} is not a valid asset name.") from None



"""Base Distance Estimators"""

# Copyright (c) 2023
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# License: BSD 3 clause

from abc import ABC, abstractmethod

import numpy as np
import numpy.typing as npt
import sklearn.base as skb


class BaseDistance(skb.BaseEstimator, ABC):
    """Base class for all distance estimators in skfolio.

    Notes
    -----
    All estimators should specify all the parameters that can be set
    at the class level in their ``__init__`` as explicit keyword
    arguments (no ``*args`` or ``**kwargs``).

    Attributes
    ----------
    codependence_ : ndarray of shape (n_assets, n_assets)
        Codependence matrix.

    distance_ : ndarray of shape (n_assets, n_assets)
        Distance matrix.
    """

    codependence_: np.ndarray
    distance_: np.ndarray

    @abstractmethod
    def __init__(self):
        pass

    @abstractmethod
    def fit(self, X: npt.ArrayLike, y=None) -> "BaseDistance":
        """Fit the Distance estimator.

        Parameters
        ----------
        X : array-like of shape (n_observations, n_assets)
            Price returns of the assets.

        y : Ignored
            Not used, present for API consistency by convention.

        Returns
        -------
        self : BaseDistance
            Fitted estimator.
        """
        pass



"""Base Portfolio module"""

# Copyright (c) 2023
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# License: BSD 3 clause

# The Portfolio class contains more than 40 measures than can be computationally
# expensive. The use of __slots__ instead of __dict__ is based on the following
# consideration:
#   * Fast Portfolio instantiation.
#   * Compute a measure only when needed.
#   * Reuse the measures functions in measures.py module independently of the
#     Portfolio class.
#   * Have the measures as Class attributes and not as Class generic
#     methods for better usability.
#   * Caching of the 40 measures.
#   * DRY by not re-writing @cached_property decorated methods for all the 40 measures.
#
# We define 7 types of attributes:
#     * Public (read and right)
#     * Private (read and right for private usage)
#     * Read-only (handled in __setattr__)
#     * Global abd local measures arguments: when they change, we clear the cache of
#       all the measures (handled in __setattr__)
#     * Attributes with custom getter and setter (using @property + private name
#       in __slots__)
#     * Attributes with custom getter without setter (read-only) that caches the result
#       (using custom decorator @cached_property_slots + private name in __slots__)
#     * Measures that are cached (handled in __getattribute__)
#
#  In order to generate the measures attributes we call the measure functions and their
#  arguments dynamically from the measures.py module. The function arguments are
#  retrieved from the class attributes following the below rules:
#     * Global measures function arguments (defined in GLOBAL_ARGS) need to be defined
#       in the class attributes with identical name.
#     * Local measures function arguments (defined in LOCAL_ARGS) need to be defined in
#       the class attributes with the argument name preceded by the measure name and
#       separated by '_'.

import warnings
from abc import abstractmethod
from typing import ClassVar

import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go

import skfolio.typing as skt
from skfolio import measures as mt
from skfolio.measures import (
    ExtraRiskMeasure,
    PerfMeasure,
    RatioMeasure,
    RiskMeasure,
)
from skfolio.utils.sorting import dominate
from skfolio.utils.tools import (
    args_names,
    cached_property_slots,
    format_measure,
)

# TODO: remove and use plotly express
pd.options.plotting.backend = "plotly"


_ZERO_THRESHOLD = 1e-5
_MEASURES = {
    e for enu in [PerfMeasure, RiskMeasure, ExtraRiskMeasure, RatioMeasure] for e in enu
}
_MEASURES_VALUES = {e.value: e for e in _MEASURES}


class BasePortfolio:
    r"""Base Portfolio class for all portfolios in skfolio.

    Parameters
    ----------
    returns : array-like of shape (n_observations,)
        Vector of portfolio returns.

    observations : array-like of shape (n_observations,)
        Vector of portfolio observations.

    name : str, optional
        Name of the portfolio.
        The default (`None`) is to use the object id.

    tag : str, optional
        Tag given to the portfolio.
        Tags are used to manipulate groups of Portfolios from a `Population`.

    fitness_measures : list[measures], optional
        List of fitness measures.
        Fitness measures are used to compute the portfolio fitness which is used to
        compute domination.
        The default (`None`) is to use the list [PerfMeasure.MEAN, RiskMeasure.VARIANCE]

    annualized_factor : float, default=252.0
        Factor used to annualize the below measures using the square-root rule:

            * Annualized Mean = Mean * factor
            * Annualized Variance = Variance * factor
            * Annualized Semi-Variance = Semi-Variance * factor
            * Annualized Standard-Deviation = Standard-Deviation * sqrt(factor)
            * Annualized Semi-Deviation = Semi-Deviation * sqrt(factor)
            * Annualized Sharpe Ratio = Sharpe Ratio * sqrt(factor)
            * Annualized Sortino Ratio = Sortino Ratio * sqrt(factor)

    risk_free_rate : float, default=0.0
        Risk-free rate. The default value is `0.0`.

    compounded : bool, default=False
        If this is set to True, cumulative returns are compounded.
        The default is `False`.

    min_acceptable_return : float, optional
        The minimum acceptable return used to distinguish "downside" and "upside"
        returns for the computation of lower partial moments:

            * First Lower Partial Moment
            * Semi-Variance
            * Semi-Deviation

        The default (`None`) is to use the mean.

    value_at_risk_beta : float, default=0.95
        The confidence level of the Portfolio VaR (Value At Risk) which represents
        the return on the worst (1-beta)% observations.
        The default value is `0.95`.

    entropic_risk_measure_theta : float, default=1.0
        The risk aversion level of the Portfolio Entropic Risk Measure.
        The default value is `1.0`.

    entropic_risk_measure_beta : float, default=0.95
        The confidence level of the Portfolio Entropic Risk Measure.
        The default value is `0.95`.

    cvar_beta : float, default=0.95
        The confidence level of the Portfolio CVaR (Conditional Value at Risk) which
        represents the expected VaR on the worst (1-beta)% observations.
        The default value is `0.95`.

    evar_beta : float, default=0.95
        The confidence level of the Portfolio EVaR (Entropic Value at Risk).
        The default value is `0.95`.

    drawdown_at_risk_beta : float, default=0.95
        The confidence level of the Portfolio Drawdown at Risk (DaR) which represents
        the drawdown on the worst (1-beta)% observations.
        The default value is `0.95`.

    cdar_beta : float, default=0.95
        The confidence level of the Portfolio CDaR (Conditional Drawdown at Risk) which
        represents the expected drawdown on the worst (1-beta)% observations.
        The default value is `0.95`.

    edar_beta : float, default=0.95
        The confidence level of the Portfolio EDaR (Entropic Drawdown at Risk).
        The default value is `0.95`.

    Attributes
    ----------
    n_observations : float
        Number of observations.

    mean : float
        Mean of the portfolio returns.

    annualized_mean : float
        Mean annualized by :math:`mean \times annualization\_factor`

    mean_absolute_deviation : float
        Mean Absolute Deviation. The deviation is the difference between the
        return and a minimum acceptable return (`min_acceptable_return`).

    first_lower_partial_moment : float
        First Lower Partial Moment. The First Lower Partial Moment is the mean of the
        returns below a minimum acceptable return (`min_acceptable_return`).

    variance : float
        Variance (Second Moment)

    annualized_variance : float
        Variance annualized by :math:`variance \times annualization\_factor`

    semi_variance : float
        Semi-variance (Second Lower Partial Moment).
        The semi-variance is the variance of the returns below a minimum acceptable
        return (`min_acceptable_return`).

    annualized_semi_variance : float
        Semi-variance annualized by
        :math:`semi\_variance \times annualization\_factor`

    standard_deviation : float
        Standard Deviation (Square Root of the Second Moment).

    annualized_standard_deviation : float
        Standard Deviation annualized by
        :math:`standard\_deviation \times \sqrt{annualization\_factor}`

    semi_deviation : float
        Semi-deviation (Square Root of the Second Lower Partial Moment).
        The Semi Standard Deviation is the Standard Deviation of the returns below a
        minimum acceptable return (`min_acceptable_return`).

    annualized_semi_deviation : float
        Semi-deviation annualized by
        :math:`semi\_deviation \times \sqrt{annualization\_factor}`

    skew : float
        Skew. The Skew is a measure of the lopsidedness of the distribution.
        A symmetric distribution have a Skew of zero.
        Higher Skew corresponds to longer right tail.

    kurtosis : float
        Kurtosis. It is a measure of the heaviness of the tail of the distribution.
        Higher Kurtosis corresponds to greater extremity of deviations (fat tails).

    fourth_central_moment : float
       Fourth Central Moment.

    fourth_lower_partial_moment : float
        Fourth Lower Partial Moment. It is a measure of the heaviness of the downside
        tail of the returns below a minimum acceptable return (`min_acceptable_return`).
        Higher Fourth Lower Partial Moment corresponds to greater extremity of downside
        deviations (downside fat tail).

    worst_realization : float
        Worst Realization which is the worst return.

    value_at_risk : float
        Historical VaR (Value at Risk).
        The VaR is the maximum loss at a given confidence level (`value_at_risk_beta`).

    cvar : float
        Historical CVaR (Conditional Value at Risk). The CVaR (or Tail VaR) represents
        the mean shortfall at a specified confidence level (`cvar_beta`).

    entropic_risk_measure : float
        Historical Entropic Risk Measure. It is a risk measure which depends on the
        risk aversion defined by the investor (`entropic_risk_measure_theta`) through
        the exponential utility function at a given confidence level
        (`entropic_risk_measure_beta`).

    evar : float
         Historical EVaR (Entropic Value at Risk). It is a coherent risk measure which
         is an upper bound for the VaR and the CVaR, obtained from the Chernoff
         inequality at a given confidence level (`evar_beta`). The EVaR can be
         represented by using the concept of relative entropy.

    drawdown_at_risk : float
        Historical Drawdown at Risk. It is the maximum drawdown at a given
        confidence level (`drawdown_at_risk_beta`).

    cdar : float
        Historical CDaR (Conditional Drawdown at Risk) at a given confidence level
        (`cdar_beta`).

    max_drawdown : float
        Maximum Drawdown.

    average_drawdown : float
        Average Drawdown.

    edar : float
        EDaR (Entropic Drawdown at Risk). It is a coherent risk measure which is an
        upper bound for the Drawdown at Risk and the CDaR, obtained from the Chernoff
        inequality at a given confidence level (`edar_beta`). The EDaR can be
        represented by using the concept of relative entropy.

    ulcer_index : float
        Ulcer Index

    gini_mean_difference : float
        Gini Mean Difference (GMD). It is the expected absolute difference between two
        realizations. The GMD is a superior measure of variability  for non-normal
        distribution than the variance. It can be used to form necessary conditions
        for second-degree stochastic dominance, while the variance cannot.

    mean_absolute_deviation_ratio : float
        Mean Absolute Deviation ratio.
        It is the excess mean (mean - risk_free_rate) divided by the MaD.

    first_lower_partial_moment_ratio : float
        First Lower Partial Moment ratio.
        It is the excess mean (mean - risk_free_rate) divided by the First Lower
        Partial Moment.

    sharpe_ratio : float
        Sharpe ratio.
        It is the excess mean (mean - risk_free_rate) divided by the standard-deviation.

    annualized_sharpe_ratio : float
        Sharpe ratio annualized by
        :math:`sharpe\_ratio \times \sqrt{annualization\_factor}`.

    sortino_ratio : float
        Sortino ratio.
        It is the excess mean (mean - risk_free_rate) divided by the semi
        standard-deviation.

    annualized_sortino_ratio : float
        Sortino ratio annualized by
        :math:`sortino\_ratio \times \sqrt{annualization\_factor}`.

    value_at_risk_ratio : float
        VaR ratio.
        It is the excess mean (mean - risk_free_rate) divided by the Value at Risk
        (VaR).

    cvar_ratio : float
        CVaR ratio.
        It is the excess mean (mean - risk_free_rate) divided by the Conditional Value
        at Risk (CVaR).

    entropic_risk_measure_ratio : float
        Entropic risk measure ratio.
        It is the excess mean (mean - risk_free_rate) divided by the Entropic risk
        measure.

    evar_ratio : float
        EVaR ratio.
        It is the excess mean (mean - risk_free_rate) divided by the EVaR (Entropic
        Value at Risk).

    worst_realization_ratio : float
        Worst Realization ratio.
        It is the excess mean (mean - risk_free_rate) divided by the Worst Realization
        (worst return).

    drawdown_at_risk_ratio : float
        Drawdown at Risk ratio.
        It is the excess mean (mean - risk_free_rate) divided by the drawdown at
        risk.

    cdar_ratio : float
        CDaR ratio.
        It is the excess mean (mean - risk_free_rate) divided by the CDaR (conditional
        drawdown at risk).

    calmar_ratio : float
        Calmar ratio.
        It is the excess mean (mean - risk_free_rate) divided by the Maximum Drawdown.

    average_drawdown_ratio : float
        Average Drawdown ratio.
        It is the excess mean (mean - risk_free_rate) divided by the Average Drawdown.

    edar_ratio : float
        EDaR ratio.
        It is the excess mean (mean - risk_free_rate) divided by the EDaR (Entropic
        Drawdown at Risk).

    ulcer_index_ratio : float
        Ulcer Index ratio.
        It is the excess mean (mean - risk_free_rate) divided by the Ulcer Index.

    gini_mean_difference_ratio : float
        Gini Mean Difference ratio.
        It is the excess mean (mean - risk_free_rate) divided by the Gini Mean
        Difference.
    """

    _read_only_attrs: ClassVar[set] = {
        "returns",
        "observations",
    }

    # Arguments globally used in measures computation
    _measure_global_args: ClassVar[set] = {
        "returns",
        "cumulative_returns",
        "drawdowns",
        "min_acceptable_return",
        "compounded",
        "risk_free_rate",
    }

    # Arguments locally used in measures computation
    _measure_local_args: ClassVar[set] = {
        "value_at_risk_beta",
        "cvar_beta",
        "entropic_risk_measure_theta",
        "entropic_risk_measure_beta",
        "evar_beta",
        "drawdown_at_risk_beta",
        "cdar_beta",
        "edar_beta",
    }

    __slots__ = {
        # public
        "tag",
        "name",
        # public read-only
        "returns",
        "observations",
        # private
        "_loaded",
        # custom getter and setter
        "_fitness_measures",
        "_annualized_factor",
        # custom getter (read-only and cached)
        "_fitness",
        "_cumulative_returns",
        "_drawdowns",
        # global args
        "min_acceptable_return",
        "compounded",
        "risk_free_rate",
        # local args
        "value_at_risk_beta",
        "cvar_beta",
        "entropic_risk_measure_theta",
        "entropic_risk_measure_beta",
        "evar_beta",
        "drawdown_at_risk_beta",
        "cdar_beta",
        "edar_beta",
        # measures
        # perf
        "mean",
        # annualized
        "annualized_mean",
        # risk measure
        "mean_absolute_deviation",
        "first_lower_partial_moment",
        "variance",
        "standard_deviation",
        "semi_variance",
        "semi_deviation",
        "fourth_central_moment",
        "fourth_lower_partial_moment",
        "value_at_risk",
        "cvar",
        "entropic_risk_measure",
        "evar",
        "worst_realization",
        "drawdown_at_risk",
        "cdar",
        "max_drawdown",
        "average_drawdown",
        "edar",
        "ulcer_index",
        "gini_mean_difference",
        "skew",
        "kurtosis",
        # annualized
        "annualized_variance",
        "annualized_semi_variance",
        "annualized_standard_deviation",
        "annualized_semi_deviation",
        # ratio
        "mean_absolute_deviation_ratio",
        "first_lower_partial_moment_ratio",
        "sharpe_ratio",
        "sortino_ratio",
        "value_at_risk_ratio",
        "cvar_ratio",
        "entropic_risk_measure_ratio",
        "evar_ratio",
        "worst_realization_ratio",
        "drawdown_at_risk_ratio",
        "cdar_ratio",
        "calmar_ratio",
        "average_drawdown_ratio",
        "edar_ratio",
        "ulcer_index_ratio",
        "gini_mean_difference_ratio",
        # annualized
        "annualized_sharpe_ratio",
        "annualized_sortino_ratio",
    }

    def __init__(
        self,
        returns: np.ndarray | list,
        observations: np.ndarray | list,
        name: str | None = None,
        tag: str | None = None,
        annualized_factor: float = 252.0,
        fitness_measures: list[skt.Measure] | None = None,
        risk_free_rate: float = 0.0,
        compounded: bool = False,
        min_acceptable_return: float | None = None,
        value_at_risk_beta: float = 0.95,
        entropic_risk_measure_theta: float = 1.0,
        entropic_risk_measure_beta: float = 0.95,
        cvar_beta: float = 0.95,
        evar_beta: float = 0.95,
        drawdown_at_risk_beta: float = 0.95,
        cdar_beta: float = 0.95,
        edar_beta: float = 0.95,
    ):
        self._loaded = False
        self._annualized_factor = annualized_factor
        self.returns = np.asarray(returns)
        self.observations = np.asarray(observations)
        self.risk_free_rate = risk_free_rate
        self.tag = tag
        self.compounded = compounded
        self.min_acceptable_return = min_acceptable_return
        self.value_at_risk_beta = value_at_risk_beta
        self.entropic_risk_measure_theta = entropic_risk_measure_theta
        self.entropic_risk_measure_beta = entropic_risk_measure_beta
        self.cvar_beta = cvar_beta
        self.evar_beta = evar_beta
        self.drawdown_at_risk_beta = drawdown_at_risk_beta
        self.cdar_beta = cdar_beta
        self.edar_beta = edar_beta

        self.name = str(id(self)) if name is None else name
        if fitness_measures is None:
            self._fitness_measures = [PerfMeasure.MEAN, RiskMeasure.VARIANCE]
        else:
            self._fitness_measures = fitness_measures
        self._loaded = True

    def __reduce__(self):
        # For fast serialization and deserialization
        # We don't want to serialize generic slots but only init arguments
        return self.__class__, tuple(
            [getattr(self, arg) for arg in args_names(self.__init__)]
        )

    def __repr__(self) -> str:
        return f"<{type(self).__name__} {self.name}>"

    def __eq__(self, other) -> bool:
        return isinstance(other, BasePortfolio) and np.array_equal(
            self.fitness, other.fitness
        )

    def __gt__(self, other) -> bool:
        if not isinstance(other, BasePortfolio):
            raise TypeError(
                "`>` not supported between instances of `Portfolio` and"
                f" `{type(other)}`"
            )
        return self.dominates(other)

    def __ge__(self, other) -> bool:
        if not isinstance(other, BasePortfolio):
            raise TypeError(
                "`>=` not supported between instances of `Portfolio` and"
                f" `{type(other)}`"
            )
        return self.__eq__(other) or self.__gt__(other)

    def __copy__(self):
        cls = self.__class__
        result = cls.__new__(cls)
        result._loaded = False
        for attr in self._slots():
            if attr not in _MEASURES_VALUES and attr != "_loaded":
                try:
                    setattr(result, attr, getattr(self, attr))
                except AttributeError:
                    pass
        result._loaded = True
        return result

    def __getattribute__(self, name):
        try:
            return object.__getattribute__(self, name)
        except AttributeError as e:
            # The Measures are the only attributes in __slots__ that are not yet
            # assigned.
            # We assign their values dynamically the first time they are called.
            if name not in _MEASURES_VALUES:
                raise AttributeError(e) from None
            measure = _MEASURES_VALUES[name]
            value = self.get_measure(measure=measure)
            setattr(self, name, value)
            return value

    def __setattr__(self, name, value):
        if name != "_loaded" and self._loaded:
            if name in self._read_only_attrs:
                raise AttributeError(
                    f"can't set attribute '{name}' because it is read-only"
                )
            if name in self._measure_global_args or name in self._measure_local_args:
                # When an attribute in GLOBAL_ARGS or LOCAL_ARGS is set, we reset all
                # the measures
                self.clear()
        object.__setattr__(self, name, value)

    def __delattr__(self, name):
        # We only want to raise an error when the attribute doesn't exist and we don't
        # want to raise an error when it's a valid attribute that has not been assigned
        # a value.
        try:
            object.__delattr__(self, name)
        except AttributeError:
            if name not in self._slots():
                raise AttributeError(
                    f"`{type(self).__name__}` object has no attribute '{name}'"
                ) from None

    def __array__(self) -> np.ndarray:
        return self.returns

    # Private methods
    def _slots(self) -> set[str]:
        slots = set()
        for s in self.__class__.__mro__:
            slots.update(getattr(s, "__slots__", set()))
        return slots

    @property
    @abstractmethod
    def composition(self) -> pd.DataFrame:
        """DataFrame of the Portfolio composition"""
        pass

    # Custom attribute setter and getter
    @property
    def fitness_measures(self) -> list[skt.Measure]:
        """Portfolio fitness measures."""
        return self._fitness_measures

    @fitness_measures.setter
    def fitness_measures(self, value: list[skt.Measure]) -> None:
        if not isinstance(value, list) or len(value) == 0:
            raise TypeError("`fitness_measures` must be a non-empty list of Measure")
        for val in value:
            if not isinstance(
                val, PerfMeasure | RiskMeasure | ExtraRiskMeasure | RatioMeasure
            ):
                raise TypeError("`fitness_measures` must be a list of Measure")
        self._fitness_measures = value
        delattr(self, "_fitness")

    @property
    def annualized_factor(self) -> float:
        """Portfolio annualized factor."""
        return self._annualized_factor

    @annualized_factor.setter
    def annualized_factor(self, value: float) -> None:
        self._annualized_factor = value
        self.clear()

    # Custom attribute getter (read-only and cached)
    @cached_property_slots
    def fitness(self) -> np.ndarray:
        """The Portfolio fitness."""
        res = []
        for measure in self.fitness_measures:
            if isinstance(measure, PerfMeasure | RatioMeasure):
                sign = 1
            else:
                sign = -1
            res.append(sign * getattr(self, str(measure.value)))
        return np.array(res)

    @cached_property_slots
    def cumulative_returns(self) -> np.ndarray:
        """Portfolio cumulative returns array."""
        return mt.get_cumulative_returns(
            returns=self.returns, compounded=self.compounded
        )

    @cached_property_slots
    def drawdowns(self) -> np.ndarray:
        """Portfolio drawdowns array."""
        return mt.get_drawdowns(returns=self.returns, compounded=self.compounded)

    # Classic property
    @property
    def n_observations(self) -> int:
        """Number of observations"""
        return len(self.observations)

    @property
    def returns_df(self) -> pd.Series:
        """Portfolio returns DataFrame."""
        return pd.Series(index=self.observations, data=self.returns, name="returns")

    @property
    def cumulative_returns_df(self) -> pd.Series:
        """Portfolio cumulative returns Series."""
        return pd.Series(
            index=self.observations,
            data=self.cumulative_returns,
            name="cumulative_returns",
        )

    @property
    def measures_df(self) -> pd.DataFrame:
        """DataFrame of all measures."""
        idx = [e.value for enu in [PerfMeasure, RiskMeasure, RatioMeasure] for e in enu]
        res = [getattr(self, attr) for attr in idx]
        return pd.DataFrame(res, index=idx, columns=["measures"])

    # Public methods
    def copy(self):
        """Copy the Portfolio attributes without its measures values."""
        return self.__copy__()

    def clear(self) -> None:
        """Clear all measures, fitness, cumulative returns and drawdowns in slots"""
        attrs = ["_fitness", "_cumulative_returns", "_drawdowns"]
        for attr in attrs + list(_MEASURES_VALUES):
            delattr(self, attr)

    def get_measure(self, measure: skt.Measure) -> float:
        """Returns the value of a given measure.

        Parameters
        ----------
        measure : PerfMeasure | RiskMeasure | ExtraRiskMeasure | RatioMeasure
            The input measure.

        Returns
        -------
        value : float
            The measure value.
        """
        if isinstance(measure, PerfMeasure | RiskMeasure | ExtraRiskMeasure):
            # We call the measure functions and their arguments dynamically.
            # The measure functions are called from the "measures" module.
            # The function arguments are retrieved from the class attributes following
            # the below rules:
            # Global measures function arguments (defined in GLOBAL_ARGS) need to be
            # defined in the class attributes with identical name.
            # Local measures function arguments need to be defined in the class
            # attributes with the argument name preceded by the measure name and
            # separated by "_".
            if measure.is_annualized:
                func = getattr(mt, str(measure.non_annualized_measure.value))
            else:
                func = getattr(mt, str(measure.value))

            args = {
                arg: (
                    getattr(self, arg)
                    if arg in self._measure_global_args
                    else getattr(self, f"{measure.value}_{arg}")
                )
                for arg in args_names(func)
            }
            try:
                value = func(**args)
                if measure in [
                    PerfMeasure.ANNUALIZED_MEAN,
                    RiskMeasure.ANNUALIZED_VARIANCE,
                    RiskMeasure.ANNUALIZED_SEMI_VARIANCE,
                ]:
                    value *= self.annualized_factor
                elif measure in [
                    RiskMeasure.ANNUALIZED_STANDARD_DEVIATION,
                    RiskMeasure.ANNUALIZED_SEMI_DEVIATION,
                ]:
                    value *= np.sqrt(self.annualized_factor)
            except Exception as e:
                warnings.warn(
                    f"Unable to calculate the portfolio '{measure.value}' with"
                    f" error: {e}",
                    stacklevel=2,
                )
                value = np.nan
        elif isinstance(measure, RatioMeasure):
            # ratio
            if measure.is_annualized:
                mean = self.annualized_mean
            else:
                mean = self.mean
            risk = getattr(self, str(measure.linked_risk_measure.value))
            value = (mean - self.risk_free_rate) / risk
        else:
            raise ValueError(f"{measure} is not a Measure.")
        return value

    def dominates(
        self, other: "BasePortfolio", idx: slice | np.ndarray | None = None
    ) -> bool:
        """Portfolio domination.

        Returns true if each objective of the current portfolio fitness is not
        strictly worse than the corresponding objective of the other portfolio fitness
        and at least one objective is strictly better.

        Parameters
        ----------
        other : BasePortfolio
            The other portfolio.

        idx : slice | array, optional
            Indexes or slice indicating on which objectives the domination is performed.
            The default (`None`) is to use all objectives.

        Returns
        -------
        value : bool
            Returns True if the Portfolio dominates the other one.
        """
        if idx is None:
            idx = slice(None)
        return dominate(self.fitness[idx], other.fitness[idx])

    def rolling_measure(
        self, measure: skt.Measure = RatioMeasure.SHARPE_RATIO, window: int = 30
    ) -> pd.Series:
        """Compute the measure over a rolling window.

        Parameters
        ----------
        measure : ct.Measure, default=RatioMeasure.SHARPE_RATIO
            The measure. The default measure is the Sharpe Ratio.

        window : int, default=30
            The window size. The default value is `30`.

        Returns
        -------
        series : pandas Series
            The rolling measure Series.
        """
        if measure.is_annualized:
            non_annualized_measure = measure.non_annualized_measure
        else:
            non_annualized_measure = measure

        if measure.is_perf:
            perf_measure = non_annualized_measure
            risk_measure = None
        elif measure.is_ratio:
            perf_measure = PerfMeasure.MEAN
            risk_measure = non_annualized_measure.linked_risk_measure
        else:
            perf_measure = None
            risk_measure = non_annualized_measure

        if risk_measure is not None:
            risk_func = getattr(mt, str(risk_measure.value))
            risk_func_args = {
                arg: (
                    getattr(self, arg)
                    if arg in self._measure_global_args
                    else getattr(self, f"{risk_measure.value}_{arg}")
                )
                for arg in args_names(risk_func)
            }

            if "drawdowns" in risk_func_args:
                del risk_func_args["drawdowns"]

                def meta_risk_func(returns):
                    drawdowns = mt.get_drawdowns(returns, compounded=self.compounded)
                    return risk_func(drawdowns=drawdowns, **risk_func_args)

            else:
                del risk_func_args["returns"]

                def meta_risk_func(returns):
                    return risk_func(returns=returns, **risk_func_args)

            if perf_measure is not None:
                perf_func = getattr(mt, str(perf_measure.value))

                def func(returns):
                    return (perf_func(returns) - self.risk_free_rate) / meta_risk_func(
                        returns
                    )

            else:
                func = meta_risk_func
        else:
            perf_func = getattr(mt, str(perf_measure.value))

            def func(returns):
                return perf_func(returns)

        rolling = (
            pd.Series(self.returns, index=self.observations)
            .rolling(window=window)
            .apply(func)
        )
        if measure.is_annualized:
            if measure in [
                PerfMeasure.ANNUALIZED_MEAN,
                RiskMeasure.ANNUALIZED_VARIANCE,
                RiskMeasure.ANNUALIZED_SEMI_VARIANCE,
            ]:
                rolling *= self.annualized_factor
            elif measure in [
                RiskMeasure.ANNUALIZED_STANDARD_DEVIATION,
                RiskMeasure.ANNUALIZED_SEMI_DEVIATION,
                RatioMeasure.ANNUALIZED_SHARPE_RATIO,
                RatioMeasure.ANNUALIZED_SORTINO_RATIO,
            ]:
                rolling *= np.sqrt(self.annualized_factor)
        return rolling

    def summary(self, formatted: bool = True) -> pd.Series:
        """Portfolio summary of all its measures.

        Parameters
        ----------
        formatted : bool, default=True
            If this is set to True, the measures are formatted into rounded string
            with units.

        Returns
        -------
        summary : pandas Series
            The Portfolio summary.
        """
        measures = (
            e
            for enu in [PerfMeasure, RiskMeasure, ExtraRiskMeasure, RatioMeasure]
            for e in enu
        )
        summary = {}
        for e in measures:
            e: skt.Measure
            try:
                if e.is_ratio:
                    base_measure = e.linked_risk_measure
                else:
                    base_measure = e
                beta = getattr(self, f"{base_measure.value}_beta")
                key = f"{e!s} at {beta:.0%}"
            except AttributeError:
                key = str(e)
            if isinstance(e, RatioMeasure) or e in [
                ExtraRiskMeasure.ENTROPIC_RISK_MEASURE,
                RiskMeasure.ULCER_INDEX,
                ExtraRiskMeasure.SKEW,
                ExtraRiskMeasure.KURTOSIS,
            ]:
                percent = False
            else:
                percent = True
            if formatted:
                value = format_measure(getattr(self, str(e.value)), percent=percent)
            else:
                value = getattr(self, str(e.value))
            summary[key] = value
        return pd.Series(summary)

    def plot_cumulative_returns(
        self, log_scale: bool = False, idx: slice | np.ndarray | None = None
    ) -> go.Figure:
        """Plot the Portfolio cumulative returns.
        Non-compounded cumulative returns start at 0.
        Compounded cumulative returns are rescaled to start at 1000.

        Parameters
        ----------
        log_scale : bool, default=False
            If this is set to True, the cumulative returns are displayed with a
            logarithm scale on the y-axis and rebased at 1000. The cumulative returns
            must be compounded otherwise an exception is raised.

        idx : slice | array, optional
            Indexes or slice of the observations to plot.
            The default (`None`) is to plot all observations.

        Returns
        -------
        plot : Figure
            Returns the plot Figure object.
        """
        if idx is None:
            idx = slice(None)
        df = self.cumulative_returns_df.iloc[idx]
        title = "Cumulative Returns"
        if self.compounded:
            yaxis_title = f"{title} (rebased at 1000)"
            if log_scale:
                title = f"{title} (compounded & log scaled)"
            else:
                title = f"{title} (compounded)"
        else:
            if log_scale:
                raise ValueError(
                    "Plotting with logarithm scaling must be done on cumulative "
                    "returns that are compounded as opposed to non-compounded."
                    "You can change to compounded with `compounded=True`"
                )
            yaxis_title = title
            title = f"{title} (non-compounded)"

        fig = df.plot()
        fig.update_layout(
            title=title,
            xaxis_title="Observations",
            yaxis_title=yaxis_title,
            showlegend=False,
        )
        if self.compounded:
            fig.update_yaxes(tickformat=".0f")
        else:
            fig.update_yaxes(tickformat=".2%")
        if log_scale:
            fig.update_yaxes(type="log")
        return fig

    def plot_returns(self, idx: slice | np.ndarray | None = None) -> go.Figure:
        """Plot the Portfolio returns

        Parameters
        ----------
        idx : slice | array, optional
            Indexes or slice of the observations to plot.
            The default (`None`) is to plot all observations.

        Returns
        -------
        plot : Figure
            Returns the plot Figure object
        """
        if idx is None:
            idx = slice(None)
        fig = self.returns_df.iloc[idx].plot()
        fig.update_layout(
            title="Returns",
            xaxis_title="Observations",
            yaxis_title="Returns",
            showlegend=False,
        )
        return fig

    def plot_rolling_measure(
        self,
        measure: skt.Measure = RatioMeasure.SHARPE_RATIO,
        window: int = 30,
    ) -> go.Figure:
        """Plot the measure over a rolling window.

        Parameters
        ----------
        measure : ct.Measure, default = RatioMeasure.SHARPE_RATIO
           The measure.

        window : int, default=30
           The window size.

        Returns
        -------
        plot : Figure
            Returns the plot Figure object
        """
        rolling = self.rolling_measure(measure=measure, window=window)
        rolling.name = f"{measure} {window} observations"
        fig = rolling.plot()
        fig.add_hline(
            y=getattr(self, measure.value),
            line_width=1,
            line_dash="dash",
            line_color="blue",
        )
        max_val = rolling.max()
        min_val = rolling.min()
        if max_val > 0:
            fig.add_hrect(
                y0=0, y1=max_val * 1.3, line_width=0, fillcolor="green", opacity=0.1
            )
        if min_val < 0:
            fig.add_hrect(
                y0=min_val * 1.3, y1=0, line_width=0, fillcolor="red", opacity=0.1
            )

        fig.update_layout(
            title=f"rolling {measure} - {window} observations window",
            xaxis_title="Observations",
            yaxis_title=str(measure),
            showlegend=False,
        )
        return fig

    def plot_composition(self) -> go.Figure:
        """Plot the Portfolio composition.

        Returns
        -------
        plot : Figure
            Returns the plot Figure object.
        """
        df = self.composition.T
        fig = px.bar(df, x=df.index, y=df.columns)
        fig.update_layout(
            title="Portfolio Composition",
            xaxis_title="Portfolio",
            yaxis_title="Weight",
            legend_title_text="Assets",
        )
        return fig



from skfolio.uncertainty_set._base import (
    BaseCovarianceUncertaintySet,
    BaseMuUncertaintySet,
    UncertaintySet,
)
from skfolio.uncertainty_set._bootstrap import (
    BootstrapCovarianceUncertaintySet,
    BootstrapMuUncertaintySet,
)
from skfolio.uncertainty_set._empirical import (
    EmpiricalCovarianceUncertaintySet,
    EmpiricalMuUncertaintySet,
)

__all__ = [
    "UncertaintySet",
    "BaseMuUncertaintySet",
    "BaseCovarianceUncertaintySet",
    "EmpiricalMuUncertaintySet",
    "EmpiricalCovarianceUncertaintySet",
    "BootstrapMuUncertaintySet",
    "BootstrapCovarianceUncertaintySet",
]



"""Scorer module"""

# Copyright (c) 2023
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# License: BSD 3 clause
# Implementation derived from:
# scikit-portfolio, Copyright (c) 2022, Carlo Nicolini, Licensed under MIT Licence.
# scikit-learn, Copyright (c) 2007-2010 David Cournapeau, Fabian Pedregosa, Olivier
# Grisel Licensed under BSD 3 clause.

from collections.abc import Callable

import numpy.typing as npt

import skfolio.typing as skt
from skfolio.optimization import BaseOptimization
from skfolio.portfolio import Portfolio


class _PortfolioScorer:
    """Portfolio Scorer wrapper"""

    def __init__(self, score_func: Callable, sign: int, kwargs: dict):
        self._score_func = score_func
        self._kwargs = kwargs
        self._sign = sign

    def __repr__(self) -> str:
        """String representation of the `PortfolioScorer`."""
        kwargs_string = "".join([f", {k}={v}" for k, v in self._kwargs.items()])
        return (
            f"make_scorer({self._score_func.__name__}"
            f"{'' if self._sign > 0 else ', greater_is_better=False'}"
            f"{kwargs_string})"
        )

    def __call__(self, estimator: BaseOptimization, X: npt.ArrayLike) -> float:
        """Compute the score of the estimator prediction on X.

        Parameters
        ----------
        estimator : BaseOptimization
            Trained estimator to use for scoring.

        X : array-like of shape (n_observations, n_assets)
            Test data that will be fed to estimator.predict.

        Returns
        -------
        score : float
            Score of the estimator prediction on X.
        """
        pred = estimator.predict(X)
        return self._sign * self._score_func(pred, **self._kwargs)


def make_scorer(
    score_func: skt.Measure | Callable,
    greater_is_better: bool | None = None,
    **kwargs,
) -> Callable:
    """Make a scorer from a :ref:`measure <measures_ref>` or from a custom score
    function.

    This is a modified version from `scikit-learn` `make_scorer` for enhanced
    functionalities with `Portfolio` objects.

    This factory function wraps scoring functions for use in
    `sklearn.model_selection.GridSearchCV` and
    `sklearn.model_selection.cross_val_score`.

    Parameters
    ----------
    score_func : Measure | callable
        If `score_func` is a :ref:`measure <measures_ref>`, we return the measure of
        the predicted :class:`~skfolio.portfolio.Portfolio` times `1` or `-1`
        depending on the `greater_is_better` parameter.

        Otherwise, `score_func` must be a score function (or loss function) with
        signature `score_func(pred, **kwargs)`. The argument `pred` is the predicted
        :class:`~skfolio.portfolio.Portfolio`.

        Note that you can convert this portfolio object into a numpy array of price
        returns with `np.asarray(pred)`.

    greater_is_better : bool, optional
        If this is set to True, `score_func` is a score function (default) meaning high
        is good, otherwise it is a loss function, meaning low is good.
        In the latter case, the scorer object will sign-flip the outcome of the `score_func`.
        The default (`None`) is to use:

        * If `score_func` is a :ref:`measure <measures_ref>`:

            * True for `PerfMeasure` and `RationMeasure`
            * False for `RiskMeasure` and `ExtraRiskMeasure`.

        * Otherwise, True.

    **kwargs : additional arguments
        Additional parameters to be passed to score_func.

    Returns
    -------
    scorer : callable
        Callable object that returns a scalar score.
    """
    if callable(score_func):
        if greater_is_better is None:
            greater_is_better = True

    else:
        measure = score_func
        if not isinstance(measure, skt.Measure):
            raise TypeError("`score_func` must be a callable or a measure")
        if greater_is_better is None:
            if measure.is_perf or measure.is_ratio:
                greater_is_better = True
            else:
                greater_is_better = False

        def score_func(pred: Portfolio) -> float:
            """Score function"""
            return getattr(pred, measure.value)

    sign = 1 if greater_is_better else -1
    return _PortfolioScorer(score_func, sign, kwargs)



"""Multi Period Portfolio module.
`MultiPeriodPortfolio` is returned by the `predict` method of Optimization estimators.
`MultiPeriodPortfolio` is a list of `Portfolio`.
"""

# Copyright (c) 2023
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# License: BSD 3 clause

import numbers
from collections.abc import Iterator

import numpy as np
import pandas as pd

import skfolio.typing as skt
from skfolio.portfolio._base import BasePortfolio
from skfolio.portfolio._portfolio import Portfolio
from skfolio.utils.tools import deduplicate_names

pd.options.plotting.backend = "plotly"


class MultiPeriodPortfolio(BasePortfolio):
    r"""Multi-Period Portfolio class.

    A Multi-Period Portfolio is composed of a list of :class:`Portfolio`.

    Parameters
    ----------
    portfolios : list[Portfolio], optional
       A list of :class:`Portfolio`. The default (`None`) is to initialize with an
       empty list.

    name : str, optional
        Name of the multi-period portfolio.
        The default (`None`) is to use the object id.

    tag : str, optional
        Tag given to the multi-period portfolio.
        Tags are used to manipulate groups of portfolios from a `Population`.

    fitness_measures : list[measures], optional
        List of fitness measures.
        Fitness measures are used to compute the portfolio fitness which is used to
        compute domination.
        The default (`None`) is to use the list [PerfMeasure.MEAN, RiskMeasure.VARIANCE]

    annualized_factor : float, default=252.0
        Factor used to annualize the below measures using the square-root rule:

            * Annualized Mean = Mean * factor
            * Annualized Variance = Variance * factor
            * Annualized Semi-Variance = Semi-Variance * factor
            * Annualized Standard-Deviation = Standard-Deviation * sqrt(factor)
            * Annualized Semi-Deviation = Semi-Deviation * sqrt(factor)
            * Annualized Sharpe Ratio = Sharpe Ratio * sqrt(factor)
            * Annualized Sortino Ratio = Sortino Ratio * sqrt(factor)


    risk_free_rate : float, default=0.0
        Risk-free rate. The default value is `0.0`.

    compounded : bool, default=False
        If this is set to True, cumulative returns are compounded.
        The default is `False`.

    min_acceptable_return : float, optional
        The minimum acceptable return used to distinguish "downside" and "upside"
        returns for the computation of lower partial moments:

            * First Lower Partial Moment
            * Semi-Variance
            * Semi-Deviation

        The default (`None`) is to use the mean.

    value_at_risk_beta : float, default=0.95
        The confidence level of the portfolio VaR (Value At Risk) which represents
        the return on the worst (1-beta)% observations.
        The default value is `0.95`.

    entropic_risk_measure_theta : float, default=1.0
        The risk aversion level of the portfolio Entropic Risk Measure.
        The default value is `1.0`.

    entropic_risk_measure_beta : float, default=0.95
        The confidence level of the portfolio Entropic Risk Measure.
        The default value is `0.95`.

    cvar_beta : float, default=0.95
        The confidence level of the portfolio CVaR (Conditional Value at Risk) which
        represents the expected VaR on the worst (1-beta)% observations.
        The default value is `0.95`.

    evar_beta : float, default=0.95
        The confidence level of the portfolio EVaR (Entropic Value at Risk).
        The default value is `0.95`.

    drawdown_at_risk_beta : float, default=0.95
        The confidence level of the portfolio Drawdown at Risk (DaR) which represents
        the drawdown on the worst (1-beta)% observations.
        The default value is `0.95`.

    cdar_beta : float, default=0.95
        The confidence level of the portfolio CDaR (Conditional Drawdown at Risk) which
        represents the expected drawdown on the worst (1-beta)% observations.
        The default value is `0.95`.

    edar_beta : float, default=0.95
        The confidence level of the portfolio EDaR (Entropic Drawdown at Risk).
        The default value is `0.95`.

    check_observations_order : bool, default=False
        If this is set to True, and if the list of portfolios is not chronologically
        sorted, an error is raised. The chronological order is determined by comparing
        the first and last observations of each portfolio.
        The default is `False`.

    Attributes
    ----------
    n_observations : float
        Number of observations.

    mean : float
        Mean of the portfolio returns.

    annualized_mean : float
        Mean annualized by :math:`mean \times annualization\_factor`

    mean_absolute_deviation : float
        Mean Absolute Deviation. The deviation is the difference between the
        return and a minimum acceptable return (`min_acceptable_return`).

    first_lower_partial_moment : float
        First Lower Partial Moment. The First Lower Partial Moment is the mean of the
        returns below a minimum acceptable return (`min_acceptable_return`).

    variance : float
        Variance (Second Moment)

    annualized_variance : float
        Variance annualized by :math:`variance \times annualization\_factor`

    semi_variance : float
        Semi-variance (Second Lower Partial Moment).
        The semi-variance is the variance of the returns below a minimum acceptable
        return (`min_acceptable_return`).

    annualized_semi_variance : float
        Semi-variance annualized by
        :math:`semi\_variance \times annualization\_factor`

    standard_deviation : float
        Standard Deviation (Square Root of the Second Moment).

    annualized_standard_deviation : float
        Standard Deviation annualized by
        :math:`standard\_deviation \times \sqrt{annualization\_factor}`

    semi_deviation : float
        Semi-deviation (Square Root of the Second Lower Partial Moment).
        The Semi Standard Deviation is the Standard Deviation of the returns below a
        minimum acceptable return (`min_acceptable_return`).

    annualized_semi_deviation : float
        Semi-deviation annualized by
        :math:`semi\_deviation \times \sqrt{annualization\_factor}`

    skew : float
        Skew. The Skew is a measure of the lopsidedness of the distribution.
        A symmetric distribution have a Skew of zero.
        Higher Skew corresponds to longer right tail.

    kurtosis : float
        Kurtosis. It is a measure of the heaviness of the tail of the distribution.
        Higher Kurtosis corresponds to greater extremity of deviations (fat tails).

    fourth_central_moment : float
       Fourth Central Moment.

    fourth_lower_partial_moment : float
        Fourth Lower Partial Moment. It is a measure of the heaviness of the downside
        tail of the returns below a minimum acceptable return (`min_acceptable_return`).
        Higher Fourth Lower Partial Moment corresponds to greater extremity of downside
        deviations (downside fat tail).

    worst_realization : float
        Worst Realization which is the worst return.

    value_at_risk : float
        Historical VaR (Value at Risk).
        The VaR is the maximum loss at a given confidence level (`value_at_risk_beta`).

    cvar : float
        Historical CVaR (Conditional Value at Risk). The CVaR (or Tail VaR) represents
        the mean shortfall at a specified confidence level (`cvar_beta`).

    entropic_risk_measure : float
        Historical Entropic Risk Measure. It is a risk measure which depends on the
        risk aversion defined by the investor (`entropic_risk_measure_theta`) through
        the exponential utility function at a given confidence level
        (`entropic_risk_measure_beta`).

    evar : float
         Historical EVaR (Entropic Value at Risk). It is a coherent risk measure which
         is an upper bound for the VaR and the CVaR, obtained from the Chernoff
         inequality at a given confidence level (`evar_beta`). The EVaR can be
         represented by using the concept of relative entropy.

    drawdown_at_risk : float
        Historical Drawdown at Risk. It is the maximum drawdown at a given
        confidence level (`drawdown_at_risk_beta`).

    cdar : float
        Historical CDaR (Conditional Drawdown at Risk) at a given confidence level
        (`cdar_beta`).

    max_drawdown : float
        Maximum Drawdown.

    average_drawdown : float
        Average Drawdown.

    edar : float
        EDaR (Entropic Drawdown at Risk). It is a coherent risk measure which is an
        upper bound for the Drawdown at Risk and the CDaR, obtained from the Chernoff
        inequality at a given confidence level (`edar_beta`). The EDaR can be
        represented by using the concept of relative entropy.

    ulcer_index : float
        Ulcer Index

    gini_mean_difference : float
        Gini Mean Difference (GMD). It is the expected absolute difference between two
        realizations. The GMD is a superior measure of variability  for non-normal
        distribution than the variance. It can be used to form necessary conditions
        for second-degree stochastic dominance, while the variance cannot.

    mean_absolute_deviation_ratio : float
        Mean Absolute Deviation ratio.
        It is the excess mean (mean - risk_free_rate) divided by the MaD.

    first_lower_partial_moment_ratio : float
        First Lower Partial Moment ratio.
        It is the excess mean (mean - risk_free_rate) divided by the First Lower
        Partial Moment.

    sharpe_ratio : float
        Sharpe ratio.
        It is the excess mean (mean - risk_free_rate) divided by the standard-deviation.

    annualized_sharpe_ratio : float
        Sharpe ratio annualized by
        :math:`sharpe\_ratio \times \sqrt{annualization\_factor}`.

    sortino_ratio : float
        Sortino ratio.
        It is the excess mean (mean - risk_free_rate) divided by the semi
        standard-deviation.

    annualized_sortino_ratio : float
        Sortino ratio annualized by
        :math:`sortino\_ratio \times \sqrt{annualization\_factor}`.

    value_at_risk_ratio : float
        VaR ratio.
        It is the excess mean (mean - risk_free_rate) divided by the Value at Risk
        (VaR).

    cvar_ratio : float
        CVaR ratio.
        It is the excess mean (mean - risk_free_rate) divided by the Conditional Value
        at Risk (CVaR).

    entropic_risk_measure_ratio : float
        Entropic risk measure ratio.
        It is the excess mean (mean - risk_free_rate) divided by the Entropic risk
        measure.

    evar_ratio : float
        EVaR ratio.
        It is the excess mean (mean - risk_free_rate) divided by the EVaR (Entropic
        Value at Risk).

    worst_realization_ratio : float
        Worst Realization ratio.
        It is the excess mean (mean - risk_free_rate) divided by the Worst Realization
        (worst return).

    drawdown_at_risk_ratio : float
        Drawdown at Risk ratio.
        It is the excess mean (mean - risk_free_rate) divided by the drawdown at
        risk.

    cdar_ratio : float
        CDaR ratio.
        It is the excess mean (mean - risk_free_rate) divided by the CDaR (conditional
        drawdown at risk).

    calmar_ratio : float
        Calmar ratio.
        It is the excess mean (mean - risk_free_rate) divided by the Maximum Drawdown.

    average_drawdown_ratio : float
        Average Drawdown ratio.
        It is the excess mean (mean - risk_free_rate) divided by the Average Drawdown.

    edar_ratio : float
        EDaR ratio.
        It is the excess mean (mean - risk_free_rate) divided by the EDaR (Entropic
        Drawdown at Risk).

    ulcer_index_ratio : float
        Ulcer Index ratio.
        It is the excess mean (mean - risk_free_rate) divided by the Ulcer Index.

    gini_mean_difference_ratio : float
        Gini Mean Difference ratio.
        It is the excess mean (mean - risk_free_rate) divided by the Gini Mean
        Difference.
    """

    __slots__ = {
        # read-only
        "_portfolios",
        "check_observations_order",
    }

    def __init__(
        self,
        portfolios: list[Portfolio] | None = None,
        name: str | None = None,
        tag: str | None = None,
        risk_free_rate: float = 0,
        annualized_factor: float = 252.0,
        fitness_measures: list[skt.Measure] | None = None,
        compounded: bool = False,
        min_acceptable_return: float | None = None,
        value_at_risk_beta: float = 0.95,
        entropic_risk_measure_theta: float = 1,
        entropic_risk_measure_beta: float = 0.95,
        cvar_beta: float = 0.95,
        evar_beta: float = 0.95,
        drawdown_at_risk_beta: float = 0.95,
        cdar_beta: float = 0.95,
        edar_beta: float = 0.95,
        check_observations_order: bool = False,
    ):
        super().__init__(
            returns=np.array([]),
            observations=np.array([]),
            name=name,
            tag=tag,
            risk_free_rate=risk_free_rate,
            annualized_factor=annualized_factor,
            fitness_measures=fitness_measures,
            compounded=compounded,
            min_acceptable_return=min_acceptable_return,
            value_at_risk_beta=value_at_risk_beta,
            cvar_beta=cvar_beta,
            entropic_risk_measure_theta=entropic_risk_measure_theta,
            entropic_risk_measure_beta=entropic_risk_measure_beta,
            evar_beta=evar_beta,
            drawdown_at_risk_beta=drawdown_at_risk_beta,
            cdar_beta=cdar_beta,
            edar_beta=edar_beta,
        )
        self.check_observations_order = check_observations_order
        self._set_portfolios(portfolios=portfolios)

    def __len__(self) -> int:
        return len(self.portfolios)

    def __getitem__(self, key: int | slice) -> Portfolio | list[Portfolio]:
        return self._portfolios[key]

    def __setitem__(self, key: int, value: Portfolio) -> None:
        if not isinstance(value, Portfolio):
            raise TypeError(f"Cannot set a value with type {type(value)}")
        new_portfolios = self._portfolios.copy()
        new_portfolios[key] = value
        self._set_portfolios(portfolios=new_portfolios)
        self.clear()

    def __delitem__(self, key: int) -> None:
        new_portfolios = self._portfolios.copy()
        del new_portfolios[key]
        self._set_portfolios(portfolios=new_portfolios)
        self.clear()

    def __iter__(self) -> Iterator[Portfolio]:
        return iter(self._portfolios)

    def __contains__(self, value: Portfolio) -> bool:
        if not isinstance(value, Portfolio):
            return False
        return value in self._portfolios

    def __neg__(self):
        return self.__class__(
            portfolios=[-p for p in self],
            tag=self.tag,
            fitness_measures=self.fitness_measures,
        )

    def __abs__(self):
        return self.__class__(
            portfolios=[abs(p) for p in self],
            tag=self.tag,
            fitness_measures=self.fitness_measures,
        )

    def __round__(self, n: int):
        return self.__class__(
            portfolios=[p.__round__(n) for p in self],
            tag=self.tag,
            fitness_measures=self.fitness_measures,
        )

    def __floor__(self):
        return self.__class__(
            portfolios=[np.floor(p) for p in self],
            tag=self.tag,
            fitness_measures=self.fitness_measures,
        )

    def __trunc__(self):
        return self.__class__(
            portfolios=[np.trunc(p) for p in self],
            tag=self.tag,
            fitness_measures=self.fitness_measures,
        )

    def __add__(self, other):
        if not isinstance(other, self.__class__):
            raise TypeError(
                "Cannot add a MultiPeriodPortfolio with an object of type"
                f" {type(other)}"
            )
        if len(self) != len(other):
            raise TypeError("Cannot add two MultiPeriodPortfolio of different sizes")
        return self.__class__(
            portfolios=[p1 + p2 for p1, p2 in zip(self, other, strict=True)],
            tag=self.tag,
            fitness_measures=self.fitness_measures,
        )

    def __sub__(self, other):
        if not isinstance(other, self.__class__):
            raise TypeError(
                "Cannot subtract a MultiPeriodPortfolio with an object of type"
                f" {type(other)}"
            )
        if len(self) != len(other):
            raise TypeError(
                "Cannot subtract two MultiPeriodPortfolio of different sizes"
            )
        return self.__class__(
            portfolios=[p1 - p2 for p1, p2 in zip(self, other, strict=True)],
            tag=self.tag,
            fitness_measures=self.fitness_measures,
        )

    def __mul__(self, other: numbers.Number | list[numbers.Number] | np.ndarray):
        if np.isscalar(other):
            portfolios = [p * other for p in self]
        else:
            portfolios = [p * a for p, a in zip(self, other, strict=True)]
        return self.__class__(
            portfolios=portfolios, tag=self.tag, fitness_measures=self.fitness_measures
        )

    __rmul__ = __mul__

    def __floordiv__(self, other: numbers.Number | list[numbers.Number] | np.ndarray):
        if np.isscalar(other):
            portfolios = [p // other for p in self]
        else:
            portfolios = [p // a for p, a in zip(self, other, strict=True)]
        return self.__class__(
            portfolios=portfolios, tag=self.tag, fitness_measures=self.fitness_measures
        )

    def __truediv__(self, other: numbers.Number | list[numbers.Number] | np.ndarray):
        if np.isscalar(other):
            portfolios = [p / other for p in self]
        else:
            portfolios = [p / a for p, a in zip(self, other, strict=True)]
        return self.__class__(
            portfolios=portfolios, tag=self.tag, fitness_measures=self.fitness_measures
        )

    # Private method
    def _set_portfolios(self, portfolios: list[Portfolio] | None = None) -> None:
        """Set the returns, observations and portfolios list.

        Parameters
        ----------
        portfolios : list[Portfolio], optional
            The list of Portfolios. The default (`None`) is to use an empty list.
        """
        returns = []
        observations = []
        if portfolios is None:
            portfolios = []
        if len(portfolios) != 0:
            for item in portfolios:
                if not isinstance(item, BasePortfolio | Portfolio):
                    raise TypeError(
                        "`portfolios` items must be of type `Portfolio`, got"
                        f" {type(item).__name__}"
                    )
                returns.append(item.returns)
                observations.append(item.observations)
            returns = np.concatenate(returns)
            observations = np.concatenate(observations)
            if self.check_observations_order:
                iteration = iter(portfolios)
                prev_p = next(iteration)
                while (p := next(iteration, None)) is not None:
                    if p.observations[0] <= prev_p.observations[-1]:
                        raise ValueError(
                            "Portfolios observations should not overlap:"
                            f" {p} overlapping {prev_p}"
                        )
                    prev_p = p
        self._loaded = False
        self._portfolios = portfolios
        self.returns = np.asarray(returns)
        self.observations = np.asarray(observations)
        self._loaded = True

    # Custom attribute setter and getter
    @property
    def portfolios(self) -> list[Portfolio]:
        """List of portfolios composing the mutli-period portfolio."""
        return self._portfolios

    @portfolios.setter
    def portfolios(self, value: list[Portfolio] | None = None):
        """Set the list of Portfolios and clear the attributes cache linked to the
        list of portfolios."""
        self._set_portfolios(portfolios=value)
        self.clear()

    # Classic property
    @property
    def assets(self) -> list:
        """List of assets names in each Portfolio."""
        return [p.assets for p in self]

    @property
    def composition(self) -> pd.DataFrame:
        """DataFrame of the Portfolio composition."""
        df = pd.concat([p.composition for p in self], axis=1)
        df.fillna(0, inplace=True)
        df.columns = deduplicate_names(df.columns)
        return df

    def summary(self, formatted: bool = True) -> pd.Series:
        """Portfolio summary of all its measures.

        Parameters
        ----------
        formatted : bool, default=True
            If this is set to True, the measures are formatted into rounded string with
            units.

        Returns
        -------
        summary : series
            Portfolio summary of all its measures.
        """
        df = super().summary(formatted=formatted)
        portfolios_number = len(self)
        avg_assets_per_portfolio = np.mean([p.n_assets for p in self])
        if formatted:
            portfolios_number = str(int(portfolios_number))
            avg_assets_per_portfolio = f"{avg_assets_per_portfolio:0.1f}"
        df["Portfolios Number"] = portfolios_number
        df["Avg nb of Assets per Portfolio"] = avg_assets_per_portfolio
        return df

    # Public methods
    def append(self, portfolio: Portfolio) -> None:
        """Append a Portfolio to the Portfolio list.

        Parameters
        ----------
        portfolio : Portfolio
            The Portfolio to append.
        """
        if self.check_observations_order and len(self) != 0:
            start_date = portfolio.observations[0]
            prev_last_date = self[-1].observations[-1]
            if start_date < prev_last_date:
                raise ValueError(
                    f"Portfolios observations should not overlap: {prev_last_date} ->"
                    f" {start_date} "
                )
        self._loaded = False
        self._portfolios.append(portfolio)
        if len(self.observations) == 0:
            # We don"t concatenate an empty array as we cannot know the dtype before.
            self.observations = portfolio.observations
            self.returns = portfolio.returns
        else:
            self.observations = np.concatenate(
                [self.observations, portfolio.observations], axis=0
            )
            self.returns = np.concatenate([self.returns, portfolio.returns], axis=0)
        self._loaded = True
        self.clear()



"""Datasets module."""

# Copyright (c) 2023
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# License: BSD 3 clause
# Implementation derived from:
# scikit-portfolio, Copyright (c) 2022, Carlo Nicolini, Licensed under MIT Licence.
# scikit-learn, Copyright (c) 2007-2010 David Cournapeau, Fabian Pedregosa, Olivier
# Grisel Licensed under BSD 3 clause.

import gzip
import os
import shutil
import urllib.request as ur
from importlib import resources
from pathlib import Path

import joblib
import pandas as pd

DATA_MODULE = "skfolio.datasets.data"


def get_data_home(data_home: str | Path | None = None) -> str:
    """Return the path of the skfolio data directory.

    This folder is used by some large dataset loaders to avoid downloading the
    data several times.

    By default, the data directory is set to a folder named 'skfolio_data' in the
    user home folder.

    Alternatively, it can be set by the 'SKFOLIO_DATA' environment
    variable or programmatically by giving an explicit folder path. The '~'
    symbol is expanded to the user home folder.

    If the folder does not already exist, it is automatically created.

    Parameters
    ----------
    data_home : str, optional
        The path to skfolio data directory. If `None`, the default path
        is `~/skfolio_data`.

    Returns
    -------
    data_home: str or path-like, optional
        The path to skfolio data directory.
    """
    if data_home is None:
        data_home = os.environ.get("SKFOLIO_DATA", os.path.join("~", "skfolio_data"))
    data_home = os.path.expanduser(data_home)
    os.makedirs(data_home, exist_ok=True)
    return data_home


def clear_data_home(data_home: str | Path | None = None) -> None:
    """Delete all the content of the data home cache.

    Parameters
    ----------
    data_home : str or path-like, optional
        The path to scikit-learn data directory. If `None`, the default path
        is `~/skfolio_data`.
    """
    data_home = get_data_home(data_home)
    shutil.rmtree(data_home)


def load_gzip_compressed_csv_data(
    data_filename: str,
    data_module: str = DATA_MODULE,
    encoding="utf-8",
    datetime_index: bool = True,
) -> pd.DataFrame:
    """Loads gzip-compressed csv files with `importlib.resources`.

    1) Open resource file with `importlib.resources.open_binary`
    2) Decompress csv file with `gzip.open`
    3) Load decompressed data with `pd.read_csv`

    Parameters
    ----------
    data_filename : str
        Name of gzip-compressed csv file  (`'*.csv.gz'`) to be loaded from
        `data_module/data_file_name`. For example `'SPX500.csv.gz'`.

    data_module : str or module, default='skfolio.datasets.data'
        Module where data lives. The default is `'skfolio.datasets.data'`.

    encoding : str, default="utf-8"
        Name of the encoding that the gzip-decompressed file will be
        decoded with. The default is 'utf-8'.

    datetime_index: bool, default=True
        If this is set to True, the DataFrame index is converted to datetime with
        format="%Y-%m-%d".
        The default is `True`.

    Returns
    -------
    df : DataFrame of shape (n_observations, n_assets)
        DataFrame with each row representing one observation and each column
        representing the asset price of a given observation.
    """
    path = resources.files(data_module).joinpath(data_filename)
    with path.open("rb") as compressed_file:
        compressed_file = gzip.open(compressed_file, mode="rt", encoding=encoding)
        df = pd.read_csv(compressed_file, sep=",", index_col=0)
        if datetime_index:
            df.index = pd.to_datetime(df.index, format="%Y-%m-%d")
        return df


def download_dataset(
    data_filename: str,
    data_home: str | Path | None = None,
    download_if_missing: bool = True,
) -> pd.DataFrame:
    """Download and save locally a dataset from the remote GitHub dataset folder.

    Parameters
    ----------
    data_filename : str
        Name of gzip-compressed csv file  (`'*.csv.gz'`) to be loaded from a remote
        GitHub dataset folder.

    data_home : str or path-like, optional
        Specify another download and cache folder for the datasets. By default,
        all skfolio data is stored in `~/skfolio_data` sub-folders.

    download_if_missing : bool, default=True
        If False, raise an OSError if the data is not locally available
        instead of trying to download the data from the source site.
        The default is `True`.

    Returns
    -------
    df : DataFrame of shape (n_observations, n_assets)
        DataFrame with each row representing one observation and each column
        representing the asset price of a given observation.
    """
    url = (
        f"https://github.com/skfolio/skfolio-datasets/raw/main/"
        f"datasets/{data_filename}.csv.gz"
    )

    data_home = get_data_home(data_home=data_home)
    filepath = os.path.join(data_home, f"{data_filename}.pkz")

    if os.path.exists(filepath):
        return joblib.load(filepath)

    if not download_if_missing:
        raise OSError("Data not found and `download_if_missing` is False")

    archive_path = os.path.join(data_home, os.path.basename(url))
    ur.urlretrieve(url, archive_path)
    df = load_gzip_compressed_csv_data(archive_path)
    joblib.dump(df, filepath, compress=6)
    os.remove(archive_path)
    return df


def load_sp500_dataset() -> pd.DataFrame:
    """Load the prices of 20 assets from the S&P 500 Index composition.

    This dataset is composed of the daily prices of 20 assets from the S&P 500
    composition starting from 1990-01-02 up to 2022-12-28.

    The data comes from the Yahoo public API.
    The price is the adjusted close which is the closing price after adjustments for
    all applicable splits and dividend distributions.
    The adjustment uses appropriate split and dividend multipliers, adhering to
    the Center for Research in Security Prices (CRSP) standards.

    ==============   ==================
    Observations     8313
    Assets           20
    ==============   ==================

    Returns
    -------
    df : DataFrame of shape (n_observations, n_assets)
        Prices DataFrame

    Examples
    --------
    >>> from skfolio.datasets import load_sp500_dataset
    >>> prices = load_sp500_dataset()
    >>> prices.head()
                    AAPL     AMD       BAC  ...       UNH       WMT      XOM
    1990-01-02  0.332589  4.1250  11.65625  ...  0.382813  5.890625  12.5000
    1990-01-03  0.334821  4.0000  11.75000  ...  0.375000  5.890625  12.3750
    1990-01-04  0.335938  3.9375  11.50000  ...  0.371094  5.859375  12.2500
    1990-01-05  0.337054  3.8125  11.25000  ...  0.355469  5.796875  12.1875
    1990-01-08  0.339286  3.8125  11.31250  ...  0.347656  5.875000  12.3750
    """
    data_filename = "sp500_dataset.csv.gz"
    df = load_gzip_compressed_csv_data(data_filename)
    return df


def load_sp500_index() -> pd.DataFrame:
    """Load the prices of the S&P 500 Index.

    This dataset is composed of the daily prices of the S&P 500 Index starting from
    1990-01-02 up to 2022-12-28.

    The data comes from the Yahoo public API.
    The price is the adjusted close which is the closing price after adjustments for
    all applicable splits and dividend distributions.
    The adjustment uses appropriate split and dividend multipliers, adhering to
    the Center for Research in Security Prices (CRSP) standards.

    ==============   ==================
    Observations     8313
    Assets           1
    ==============   ==================

    Returns
    -------
    df : DataFrame of shape (n_observations, n_assets)
        Prices DataFrame

    Examples
    --------
    >>> from skfolio.datasets import load_sp500_index
    >>> prices = load_sp500_index()
    >>> prices.head()
                 SP500
    Date
    1990-01-02  359.69
    1990-01-03  358.76
    1990-01-04  355.67
    1990-01-05  352.20
    1990-01-08  353.79
    """
    data_filename = "sp500_index.csv.gz"
    df = load_gzip_compressed_csv_data(data_filename)
    return df


def load_factors_dataset() -> pd.DataFrame:
    """Load the prices of 5 factor ETFs.

    This dataset is composed of the daily prices of 5 ETF representing common factors
    starting from 2014-01-02 up to 2022-12-28.

    The factors are:

        * "MTUM": Momentum
        * "QUAL": Quanlity
        * "SIZE": Size
        * "VLUE": Value
        * "USMV": low volatility

    The data comes from the Yahoo public API.
    The price is the adjusted close which is the closing price after adjustments for
    all applicable splits and dividend distributions.
    The adjustment uses appropriate split and dividend multipliers, adhering to
    the Center for Research in Security Prices (CRSP) standards.

    ==============   ==================
    Observations     2264
    Assets           5
    ==============   ==================

    Returns
    -------
    df : DataFrame of shape (n_observations, n_assets)
        Prices DataFrame

    Examples
    --------
    >>> from skfolio.datasets import load_factors_dataset
    >>> prices = load_factors_dataset()
    >>> prices.head()
                  MTUM    QUAL    SIZE    USMV    VLUE
    Date
    2014-01-02  52.704  48.351  48.986  29.338  47.054
    2014-01-03  52.792  48.256  48.722  29.330  46.999
    2014-01-06  52.677  48.067  48.722  29.263  46.991
    2014-01-07  53.112  48.455  48.731  29.430  47.253
    2014-01-08  53.502  48.437  48.731  29.422  47.253
    """
    data_filename = "factors_dataset.csv.gz"
    df = load_gzip_compressed_csv_data(data_filename)
    return df


def load_ftse100_dataset(data_home=None, download_if_missing=True) -> pd.DataFrame:
    """Load the prices of 64 assets from the FTSE 100 Index composition.

    This dataset is composed of the daily prices of 64 assets from the FTSE 100 Index
    starting from 2000-01-04 up to 2023-05-31.

    The data comes from the Yahoo public API.
    The price is the adjusted close which is the closing price after adjustments for
    all applicable splits and dividend distributions.
    The adjustment uses appropriate split and dividend multipliers, adhering to
    the Center for Research in Security Prices (CRSP) standards.
    The data contains NaN.

    ==============   ==================
    Observations     5960
    Assets           64
    ==============   ==================

    Parameters
    ----------
    data_home : str, optional
        Specify another download and cache folder for the datasets.
        By default, all skfolio data is stored in `~/skfolio_data` subfolders.

    download_if_missing : bool, default=True
        If False, raise an OSError if the data is not locally available
        instead of trying to download the data from the source site.

    Returns
    -------
    df : DataFrame of shape (n_observations, n_assets)
        Prices DataFrame

    Examples
    --------
    >>> from skfolio.datasets import load_ftse100_dataset
    >>> prices = load_ftse100_dataset()
    >>> prices.head()
                  AAL.L    ABF.L   AHT.L  ANTO.L  ...   VOD.L   WEIR.L    WPP.L    WTB.L
    Date                                          ...
    2000-01-04  535.354  205.926  97.590  40.313  ...  72.562  115.240  512.249  382.907
    2000-01-05  540.039  209.185  96.729  40.313  ...  69.042  118.483  462.080  381.972
    2000-01-06  553.289  229.048  95.581  40.452  ...  66.950  124.220  458.119  386.337
    2000-01-07  572.829  222.220  95.581  40.452  ...  70.716  121.725  475.283  405.046
    2000-01-10  578.852  224.548  92.711  40.685  ...  74.285  121.476  498.254  392.885
    """
    data_filename = "ftse100_dataset"
    df = download_dataset(
        data_filename, data_home=data_home, download_if_missing=download_if_missing
    )
    return df


def load_nasdaq_dataset(data_home=None, download_if_missing=True) -> pd.DataFrame:
    """Load the prices of 1455 assets from the NASDAQ Composite Index.

    This dataset is composed of the daily prices of 1455 assets from the NASDAQ
    Composite starting from 2018-01-02 up to 2023-05-31.

    The data comes from the Yahoo public API.
    The price is the adjusted close which is the closing price after adjustments for
    all applicable splits and dividend distributions.
    The adjustment uses appropriate split and dividend multipliers, adhering to
    the Center for Research in Security Prices (CRSP) standards.

    ==============   ==================
    Observations     1362
    Assets           1455
    ==============   ==================

    Parameters
    ----------
    data_home : str, optional
        Specify another download and cache folder for the datasets.
        By default, all skfolio data is stored in `~/skfolio_data` subfolders.

    download_if_missing : bool, default=True
        If False, raise an OSError if the data is not locally available
        instead of trying to download the data from the source site.

    Returns
    -------
    df : DataFrame of shape (n_observations, n_assets)
        Prices DataFrame

    Examples
    --------
    >>> from skfolio.datasets import load_nasdaq_dataset
    >>> prices = load_nasdaq_dataset()
    >>> prices.head()
                   AAL   AAOI    AAON    AAPL  ...  ZVRA   ZYME    ZYNE   ZYXI
    Date                                       ...
    2018-01-02  51.648  37.91  35.621  41.310  ...  66.4  7.933  12.995  2.922
    2018-01-03  51.014  37.89  36.247  41.303  ...  72.8  7.965  13.460  2.913
    2018-01-04  51.336  38.38  36.103  41.495  ...  78.4  8.430  12.700  2.869
    2018-01-05  51.316  38.89  36.681  41.967  ...  77.6  8.400  12.495  2.780
    2018-01-08  50.809  38.37  36.103  41.811  ...  82.4  8.310  12.550  2.825
    """
    data_filename = "nasdaq_dataset"
    df = download_dataset(
        data_filename, data_home=data_home, download_if_missing=download_if_missing
    )
    return df



"""Preprocessing module to transform X to returns."""

# Copyright (c) 2023
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# License: BSD 3 clause

import numpy as np
import pandas as pd


def prices_to_returns(
    X: pd.DataFrame,
    y: pd.DataFrame | None = None,
    log_returns: bool = False,
    nan_threshold: float = 1,
    join: str = "outer",
) -> pd.DataFrame | tuple[pd.DataFrame, pd.DataFrame]:
    r"""Transforms a DataFrame of prices to linear or logarithmic returns.

    Linear returns (also called simple returns) are defined as:
        .. math:: \frac{S_{t}}{S_{t-1}} - 1

    Logarithmic returns (also called continuously compounded return) are defined as:
        .. math:: ln\Biggl(\frac{S_{t}}{S_{t-1}}\Biggr)

    With :math:`S_{t}` the asset price at time :math:`t`.

    .. warning::

        The linear returns aggregate across securities, meaning that the linear return
        of the portfolio is the weighted average of the linear returns of the
        securities. For this reason, **portfolio optimization should be performed
        using linear returns** [1]_.

        On the other hand, the logarithmic returns aggregate across time, meaning that
        the total logarithmic return over K time periods is the sum of all K
        single-period logarithmic returns.

    .. seealso::

        :ref:`data preparation <data_preparation>`

    Parameters
    ----------
    X : DataFrame
        The DataFrame of assets prices.

    y : DataFrame, optional
        The DataFrame of target or factors prices.
        If provided, it is joined with the DataFrame of prices to ensure identical
        observations.

    log_returns : bool, default=True
        If this is set to True, logarithmic returns are used instead of simple returns.

    join : str, default='outer
        The join method between `X` and `y` when `y` is provided.

    nan_threshold : float, default=1.0
        Drop observations (rows) that have a percentage of missing assets prices above
        this threshold. The default (`1.0`) is to keep all the observations.

    Returns
    -------
    X : DataFrame
        The DataFrame of price returns of the input `X`.

    y : DataFrame, optional
        The DataFrame of price returns of the input `y` when provided.

    References
    ----------
    .. [1]  "Linear vs. Compounded Returns - Common Pitfalls in Portfolio Management".
        GARP Risk Professional.
        Attilio Meucci (2010).
    """
    if not isinstance(X, pd.DataFrame):
        raise TypeError("`X` must be a DataFrame")

    if y is None:
        df = X.copy()
    else:
        if not isinstance(y, pd.DataFrame):
            raise TypeError("`y` must be a DataFrame")
        df = X.join(y, how="outer")

    n_observations, n_assets = X.shape

    # Remove observations with missing X above threshold
    if nan_threshold is not None:
        nan_threshold = float(nan_threshold)
        if not 0 < nan_threshold <= 1:
            raise ValueError("`nan_threshold` must be between 0 and 1")
        count_nan = df.isna().sum(axis=1)
        to_drop = count_nan[count_nan > n_assets * nan_threshold].index
        if len(to_drop) > 0:
            df.drop(to_drop, axis=0, inplace=True)

    # Forward fill missing values
    df.ffill(inplace=True)
    # Drop rows if any of its values is missing
    df.dropna(axis=0, how="any", inplace=True)
    # Drop column if all its values are missing
    df.dropna(axis=1, how="all", inplace=True)

    # returns
    all_returns = df.pct_change().dropna()
    if log_returns:
        all_returns = np.log1p(all_returns)

    if y is None:
        return all_returns
    returns = all_returns[[x for x in X.columns if x in df.columns]]
    factor_returns = all_returns[[x for x in y.columns if x in df.columns]]
    return returns, factor_returns



"""skfolio package"""

# Author: Hugo Delatte <delatte.hugo@gmail.com>
# License: BSD 3 clause
import importlib.metadata

from skfolio.measures import (
    BaseMeasure,
    ExtraRiskMeasure,
    PerfMeasure,
    RatioMeasure,
    RiskMeasure,
)
from skfolio.population import Population
from skfolio.portfolio import BasePortfolio, MultiPeriodPortfolio, Portfolio

__version__ = importlib.metadata.version("skfolio")

__all__ = [
    "BaseMeasure",
    "PerfMeasure",
    "RiskMeasure",
    "ExtraRiskMeasure",
    "RatioMeasure",
    "BasePortfolio",
    "Portfolio",
    "MultiPeriodPortfolio",
    "Population",
]



"""Base Composition estimator.
Follow same implementation as Base composition from sklearn
"""

# Copyright (c) 2023
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# License: BSD 3 clause
# Implementation derived from:
# scikit-learn, Copyright (c) 2007-2010 David Cournapeau, Fabian Pedregosa, Olivier
# Grisel Licensed under BSD 3 clause.

from abc import ABC, abstractmethod
from contextlib import suppress

import sklearn.base as skb


class BaseComposition(skb.BaseEstimator, ABC):
    """Handles parameter management for ensemble estimators."""

    @abstractmethod
    def __init__(self):
        pass

    def _get_params(self, attr, deep=True):
        out = super().get_params(deep=deep)
        if not deep:
            return out

        estimators = getattr(self, attr)
        try:
            out.update(estimators)
        except (TypeError, ValueError):
            # Ignore TypeError for cases where estimators is not a list of
            # (name, estimator) and ignore ValueError when the list is not
            # formatted correctly. This is to prevent errors when calling
            # `set_params`. `BaseEstimator.set_params` calls `get_params` which
            # can error for invalid values for `estimators`.
            return out

        for name, estimator in estimators:
            if hasattr(estimator, "get_params"):
                for key, value in estimator.get_params(deep=True).items():
                    out[f"{name}__{key}"] = value
        return out

    def _set_params(self, attr, **params):
        # Ensure strict ordering of parameter setting:
        # 1. All steps
        if attr in params:
            setattr(self, attr, params.pop(attr))
        # 2. Replace items with estimators in params
        items = getattr(self, attr)
        if isinstance(items, list) and items:
            # Get item names used to identify valid names in params
            # `zip` raises a TypeError when `items` does not contains
            # elements of length 2
            with suppress(TypeError):
                item_names, _ = zip(*items, strict=True)
                for name in params:
                    if "__" not in name and name in item_names:
                        self._replace_estimator(attr, name, params.pop(name))

        # 3. Step parameters and other initialisation arguments
        super().set_params(**params)
        return self

    def _replace_estimator(self, attr, name, new_val):
        # assumes `name` is a valid estimator name
        new_estimators = list(getattr(self, attr))
        for i, (estimator_name, _) in enumerate(new_estimators):
            if estimator_name == name:
                new_estimators[i] = (name, new_val)
                break
        setattr(self, attr, new_estimators)

    def _validate_names(self, names):
        if len(set(names)) != len(names):
            raise ValueError(f"Names provided are not unique: {list(names)!r}")
        invalid_names = set(names).intersection(self.get_params(deep=False))
        if invalid_names:
            raise ValueError(
                f"Estimator names conflict with constructor arguments: {sorted(invalid_names)!r}"
            )
        invalid_names = [name for name in names if "__" in name]
        if invalid_names:
            raise ValueError(
                f"Estimator names must not contain __: got {invalid_names!r}"
            )



"""Empirical Uncertainty Set estimators."""

# Copyright (c) 2023
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# License: BSD 3 clause
# Implementation derived from:
# Riskfolio-Lib, Copyright (c) 2020-2023, Dany Cajas, Licensed under BSD 3 clause.
# scikit-learn, Copyright (c) 2007-2010 David Cournapeau, Fabian Pedregosa, Olivier
# Grisel Licensed under BSD 3 clause.

import numpy as np
import numpy.typing as npt
import scipy.stats as st

from skfolio.prior import BasePrior, EmpiricalPrior
from skfolio.uncertainty_set._base import (
    BaseCovarianceUncertaintySet,
    BaseMuUncertaintySet,
    UncertaintySet,
)
from skfolio.utils.stats import commutation_matrix
from skfolio.utils.tools import check_estimator


class EmpiricalMuUncertaintySet(BaseMuUncertaintySet):
    r"""Empirical Mu Uncertainty Set.

    Compute the expected returns ellipsoidal uncertainty set [1]_:

    .. math:: U_{\mu}=\left\{\mu\,|\left(\mu-\hat{\mu}\right)S^{-1}\left(\mu-\hat{\mu}\right)^{T}\leq\kappa^{2}\right\}

    Under the assumption that :math:`\Sigma` is given, the distribution of the sample
    estimator :math:`\hat{\mu}` based on an i.i.d. sample
    :math:`R_{t}\sim N(\mu, \Sigma), t=1,...,T` is given by
    :math:`\hat{\mu}\sim N(\mu, \frac{1}{T}\Sigma)`.

    The size of the ellipsoid  :math:`\kappa` (confidence region), is computed using:

    .. math:: \kappa^2 = \chi^2_{n\_assets} (\beta)

    with :math:`\chi^2_{n\_assets}(\beta)` the inverse cumulative distribution function
    of the chi-squared distribution with `n_assets` degrees of freedom at the
    :math:`\beta` confidence level.

    The Shape of the ellipsoid :math:`S` is computed using:

    .. math:: S = \frac{1}{T}\Sigma

    with the option to force the non-diagonal elements of the covariance matrix to zero.

    Parameters
    ----------
    prior_estimator : BasePrior, optional
        The :ref:`prior estimator <prior>` used to estimate the assets covariance
        matrix. The default (`None`) is to use :class:`~skfolio.prior.EmpiricalPrior`.

    confidence_level : float , default=0.95
        Confidence level :math:`\beta` of the inverse cumulative distribution function
        of the chi-squared distribution. The default value is `0.95`.

    diagonal : bool, default=True
        If this is set to True, the non-diagonal elements of the covariance matrix are
        set to zero.

    Attributes
    ----------
    uncertainty_set_ : UncertaintySet
        Mu Uncertainty set :class:`~skfolio.uncertainty_set.UncertaintySet`.

    prior_estimator_ : BasePrior
        Fitted `prior_estimator`.

    References
    ----------
    .. [1]  "Robustness properties of mean-variance portfolios",
        Optimization: A Journal of Mathematical Programming and Operations Research,
        Schöttle & Werner (2009).
    """

    prior_estimator_: BasePrior

    def __init__(
        self,
        prior_estimator: BasePrior | None = None,
        confidence_level: float = 0.95,
        diagonal: bool = True,
    ):
        self.prior_estimator = prior_estimator
        self.confidence_level = confidence_level
        self.diagonal = diagonal

    def fit(
        self, X: npt.ArrayLike, y: npt.ArrayLike | None = None
    ) -> "EmpiricalMuUncertaintySet":
        """Fit the Empirical Mu Uncertainty set estimator.

        Parameters
        ----------
        X : array-like of shape (n_observations, n_assets)
            Price returns of the assets.

        y : array-like of shape (n_observations, n_factors), optional
            Price returns of factors.
            The default is `None`.

        Returns
        -------
        self : EmpiricalMuUncertaintySet
            Fitted estimator.
        """
        self.prior_estimator_ = check_estimator(
            self.prior_estimator,
            default=EmpiricalPrior(),
            check_type=BasePrior,
        )
        # fitting estimators
        self.prior_estimator_.fit(X, y)

        prior_model = self.prior_estimator_.prior_model_
        n_observations, n_assets = prior_model.returns.shape
        k = np.sqrt(st.chi2.ppf(q=self.confidence_level, df=n_assets))

        sigma = prior_model.covariance / n_observations
        if self.diagonal:
            sigma = np.diag(np.diag(sigma))

        self.uncertainty_set_ = UncertaintySet(k=k, sigma=sigma)
        return self


class EmpiricalCovarianceUncertaintySet(BaseCovarianceUncertaintySet):
    r"""Empirical Covariance Uncertainty set.

    Compute the covariance ellipsoidal uncertainty set [1]_:

    .. math:: U_{\Sigma}=\left\{\Sigma\,|\left(\text{vec}(\Sigma)-\text{vec}(\hat{\Sigma})\right)S^{-1}\left(\text{vec}(\Sigma)-\text{vec}(\hat{\Sigma})\right)^{T}\leq k^{2}\,,\,\Sigma\succeq 0\right\}

    We consider the Wishart distribution for the covariance matrix:

    .. math:: \hat{\Sigma}\sim W(\frac{1}{T-1}\Sigma, T-1)

    The size of the ellipsoid  :math:`\kappa` (confidence region), is computed using:

    .. math:: \kappa^2 = \chi^2_{n\_assets^2} (\beta)

    with :math:`\chi^2_{n\_assets^2}(\beta)` the inverse cumulative distribution
    function of the chi-squared distribution with `n_assets` degrees of freedom at the
    :math:`\beta` confidence level.

    The Shape of the ellipsoid :math:`S` is based on a closed form solution of the
    covariance matrix of the Wishart distributed random variable by using the vector
    notation :math:`vec(x)`:

    .. math:: Cov[vec(\hat{\Sigma})]=\frac{1}{T-1}(I_{n^2} + K_{nn})(\Sigma \otimes \Sigma)

    with :math:`K_{nn}` denoting a commutation matrix and :math:`\otimes` representing
    the Kronecker product.

    Parameters
    ----------
    prior_estimator : BasePrior, optional
        The :ref:`prior estimator <prior>` used to estimate the assets covariance
        matrix. The default (`None`) is to use :class:`~skfolio.prior.EmpiricalPrior`.

    confidence_level : float , default=0.95
        Confidence level :math:`\beta` of the inverse cumulative distribution function
        of the chi-squared distribution. The default value is `0.95`.

    diagonal : bool, default=True
        If this is set to True, the non-diagonal elements of the covariance matrix are
        set to zero.

    Attributes
    ----------
    uncertainty_set_ : UncertaintySet
        Covariance Uncertainty set :class:`~skfolio.uncertainty_set.UncertaintySet`.

    prior_estimator_ : BasePrior
        Fitted `prior_estimator`.

    References
    ----------
    .. [1]  "Robustness properties of mean-variance portfolios",
        Optimization: A Journal of Mathematical Programming and Operations Research,
        Schöttle & Werner (2009).
    """

    prior_estimator_: BasePrior

    def __init__(
        self,
        prior_estimator: BasePrior | None = None,
        confidence_level: float = 0.95,
        diagonal: bool = True,
    ):
        self.prior_estimator = prior_estimator
        self.confidence_level = confidence_level
        self.diagonal = diagonal

    def fit(
        self, X: npt.ArrayLike, y: npt.ArrayLike | None = None
    ) -> "EmpiricalCovarianceUncertaintySet":
        """Fit the Empirical Covariance Uncertainty set estimator.

        Parameters
        ----------
        X : array-like of shape (n_observations, n_assets)
            Price returns of the assets.

        y : array-like of shape (n_observations, n_factors), optional
            Price returns of factors.
            The default is `None`.

        Returns
        -------
        self : EmpiricalCovarianceUncertaintySet
            Fitted estimator.
        """
        self.prior_estimator_ = check_estimator(
            self.prior_estimator,
            default=EmpiricalPrior(),
            check_type=BasePrior,
        )
        # fitting estimators
        self.prior_estimator_.fit(X, y)

        prior_model = self.prior_estimator_.prior_model_
        n_observations, n_assets = prior_model.returns.shape
        k = np.sqrt(st.chi2.ppf(q=self.confidence_level, df=n_assets**2))

        sigma = prior_model.covariance / n_observations
        if self.diagonal:
            sigma = np.diag(np.diag(sigma))

        sigma = np.diag(
            np.diag(
                n_observations
                * (np.identity(n_assets**2) + commutation_matrix(sigma))
                @ np.kron(sigma, sigma)
            )
        )

        self.uncertainty_set_ = UncertaintySet(k=k, sigma=sigma)
        return self



"""Module that includes all Measures functions used across `skfolio`."""

# Author: Hugo Delatte <delatte.hugo@gmail.com>
# License: BSD 3 clause

from skfolio.measures._enums import (
    BaseMeasure,
    ExtraRiskMeasure,
    PerfMeasure,
    RatioMeasure,
    RiskMeasure,
)
from skfolio.measures._measures import (
    average_drawdown,
    cdar,
    cvar,
    drawdown_at_risk,
    edar,
    effective_number_assets,
    entropic_risk_measure,
    evar,
    first_lower_partial_moment,
    fourth_central_moment,
    fourth_lower_partial_moment,
    get_cumulative_returns,
    get_drawdowns,
    gini_mean_difference,
    kurtosis,
    max_drawdown,
    mean,
    mean_absolute_deviation,
    owa_gmd_weights,
    semi_deviation,
    semi_variance,
    skew,
    standard_deviation,
    third_central_moment,
    ulcer_index,
    value_at_risk,
    variance,
    worst_realization,
)

__all__ = [
    "BaseMeasure",
    "PerfMeasure",
    "RiskMeasure",
    "ExtraRiskMeasure",
    "RatioMeasure",
    "mean",
    "get_cumulative_returns",
    "get_drawdowns",
    "variance",
    "semi_variance",
    "standard_deviation",
    "semi_deviation",
    "third_central_moment",
    "fourth_central_moment",
    "fourth_lower_partial_moment",
    "cvar",
    "mean_absolute_deviation",
    "value_at_risk",
    "worst_realization",
    "first_lower_partial_moment",
    "entropic_risk_measure",
    "evar",
    "drawdown_at_risk",
    "cdar",
    "max_drawdown",
    "average_drawdown",
    "edar",
    "ulcer_index",
    "gini_mean_difference",
    "owa_gmd_weights",
    "skew",
    "kurtosis",
    "effective_number_assets",
]



"""Model validation module."""

# Copyright (c) 2023
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# License: BSD 3 clause
# Implementation derived from:
# scikit-portfolio, Copyright (c) 2022, Carlo Nicolini, Licensed under MIT Licence.
# scikit-learn, Copyright (c) 2007-2010 David Cournapeau, Fabian Pedregosa, Olivier
# Grisel Licensed under BSD 3 clause.

import numpy as np
import numpy.typing as npt
import sklearn as sk
import sklearn.base as skb
import sklearn.model_selection as skm
import sklearn.utils as sku
import sklearn.utils.parallel as skp

from skfolio.model_selection._combinatorial import BaseCombinatorialCV
from skfolio.population import Population
from skfolio.portfolio import MultiPeriodPortfolio
from skfolio.utils.tools import fit_and_predict, safe_split


def cross_val_predict(
    estimator: skb.BaseEstimator,
    X: npt.ArrayLike,
    y: npt.ArrayLike = None,
    groups: np.ndarray | None = None,
    cv: skm.BaseCrossValidator | BaseCombinatorialCV | int | None = None,
    n_jobs: int | None = None,
    method: str = "predict",
    verbose: int = 0,
    fit_params: dict | None = None,
    pre_dispatch: str = "2*n_jobs",
    column_indices: np.ndarray | None = None,
    portfolio_params: dict | None = None,
) -> MultiPeriodPortfolio | Population:
    """Generate cross-validated `Portfolios` estimates.

    The data is split according to the `cv` parameter.
    The optimization estimator is fitted on the training set and portfolios are
    predicted on the corresponding test set.

    For non-combinatorial cross-validation like `Kfold`, the output is the predicted
    :class:`~skfolio.portfolio.MultiPeriodPortfolio` where
    each :class:`~skfolio.portfolio.Portfolio` corresponds to the prediction on each
    train/test pair (`k` portfolios for `Kfold`).

    For combinatorial cross-validation
    like :class:`~skfolio.model_selection.CombinatorialPurgedCV`, the output is the
    predicted :class:`~skfolio.population.Population` of multiple
    :class:`~skfolio.portfolio.MultiPeriodPortfolio` (each test outputs are a
    collection of multiple paths instead of one single path).

    Parameters
    ----------
    estimator : BaseOptimization
        :ref:`Optimization estimators <optimization>` use to fit the data.

    X : array-like of shape (n_observations, n_assets)
        Price returns of the assets.

    y : array-like of shape (n_observations, n_targets), optional
        Target data (optional).
        For example, the price returns of the factors.

    groups : array-like of shape (n_observations,), optional
        Group labels for the samples used while splitting the dataset into
        train/test set. Only used in conjunction with a "Group" `cv`
        instance (e.g., `GroupKFold`).

    cv : int | cross-validation generator, optional
        Determines the cross-validation splitting strategy.
        Possible inputs for cv are:

        * None, to use the default 5-fold cross validation,
        * int, to specify the number of folds in a `(Stratified)KFold`,
        * `CV splitter`,
        * An iterable that generates (train, test) splits as arrays of indices.

    n_jobs : int, optional
        The number of jobs to run in parallel for `fit` of all `estimators`.
        `None` means 1 unless in a `joblib.parallel_backend` context. -1 means
        using all processors.

    method : str
        Invokes the passed method name of the passed estimator.

    verbose : int, default=0
        The verbosity level.

    fit_params : dict, optional
        Parameters to pass to the fit method of the estimator.

    pre_dispatch : int or str, default='2*n_jobs'
        Controls the number of jobs that get dispatched during parallel
        execution. Reducing this number can be useful to avoid an
        explosion of memory consumption when more jobs get dispatched
        than CPUs can process. This parameter can be:

            * None, in which case all the jobs are immediately
              created and spawned. Use this for lightweight and
              fast-running jobs, to avoid delays due to on-demand
              spawning of the jobs

            * An int, giving the exact number of total jobs that are
              spawned

            * A str, giving an expression as a function of n_jobs,
              as in '2*n_jobs'

    column_indices : ndarray, optional
        Indices of the `X` columns to cross-validate on.

    portfolio_params :  dict, optional
        Additional portfolio parameters passed to `MultiPeriodPortfolio`.

    Returns
    -------
    predictions : MultiPeriodPortfolio | Population
        This is the result of calling `predict`
    """
    X, y = safe_split(X, y, indices=column_indices, axis=1)
    X, y, groups = sku.indexable(X, y, groups)
    cv = skm.check_cv(cv, y)
    splits = list(cv.split(X, y, groups))
    portfolio_params = {} if portfolio_params is None else portfolio_params.copy()

    # We ensure that the folds are not shuffled
    if not isinstance(cv, BaseCombinatorialCV):
        try:
            if cv.shuffle:
                raise ValueError(
                    "`cross_val_predict` only works with cross-validation setting"
                    " `shuffle=False`"
                )
        except AttributeError:
            # If we cannot find the attribute shuffle, we check if the first folds
            # are shuffled
            for fold in splits[0]:
                if not np.all(np.diff(fold) > 0):
                    raise ValueError(
                        "`cross_val_predict` only works with un-shuffled folds"
                    ) from None

    # We clone the estimator to make sure that all the folds are independent
    # and that it is pickle-able.
    parallel = skp.Parallel(n_jobs=n_jobs, verbose=verbose, pre_dispatch=pre_dispatch)
    # TODO remove when https://github.com/joblib/joblib/issues/1071 is fixed
    predictions = parallel(
        skp.delayed(fit_and_predict)(
            sk.clone(estimator),
            X,
            y,
            train=train,
            test=test,
            fit_params=fit_params,
            method=method,
        )
        for train, test in splits
    )

    if isinstance(cv, BaseCombinatorialCV):
        path_ids = cv.get_path_ids()
        path_nb = np.max(path_ids) + 1
        portfolios = [[] for _ in range(path_nb)]
        for i, prediction in enumerate(predictions):
            for j, p in enumerate(prediction):
                path_id = path_ids[i, j]
                portfolios[path_id].append(p)
        name = portfolio_params.pop("name", "path")
        pred = Population(
            [
                MultiPeriodPortfolio(
                    name=f"{name}_{i}", portfolios=portfolios[i], **portfolio_params
                )
                for i in range(path_nb)
            ]
        )
    else:
        # We need to re-order the test folds in case they were un-ordered by the
        # CV generator.
        # Because the tests folds are not shuffled, we use the first index of each
        # fold to order them.
        test_indices = np.concatenate([test for _, test in splits])
        if np.unique(test_indices, axis=0).shape[0] != test_indices.shape[0]:
            raise ValueError(
                "`cross_val_predict` only works with non-duplicated test indices"
            )
        test_indices = [test for _, test in splits]
        sorted_fold_id = np.argsort([x[0] for x in test_indices])
        pred = MultiPeriodPortfolio(
            portfolios=[predictions[fold_id] for fold_id in sorted_fold_id],
            check_observations_order=False,
            **portfolio_params,
        )

    return pred



"""Base Covariance Estimators."""

# Copyright (c) 2023
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# License: BSD 3 clause
# Implementation derived from:
# scikit-learn, Copyright (c) 2007-2010 David Cournapeau, Fabian Pedregosa, Olivier
# Grisel Licensed under BSD 3 clause.

from abc import ABC, abstractmethod

import numpy as np
import numpy.typing as npt
import sklearn.base as skb

from skfolio.exceptions import NonPositiveVarianceError
from skfolio.utils.stats import cov_nearest


class BaseCovariance(skb.BaseEstimator, ABC):
    """Base class for all covariance estimators in `skfolio`.

    Parameters
    ----------
    nearest : bool, default=False
        If this is set to True, the covariance is replaced by the nearest covariance
        matrix that is positive definite and with a Cholesky decomposition than can be
        computed. The variance is left unchanged. A covariance matrix is in theory PSD.
        However, due to floating-point inaccuracies, we can end up with a covariance
        matrix that is slightly non-PSD or where Cholesky decomposition is failing.
        This often occurs in high dimensional problems.
        For more details, see :func:`~skfolio.units.stats.cov_nearest`.
        The default is `False`.

    higham : bool, default=False
        If this is set to True, the Higham & Nick (2002) algorithm is used to find the
        nearest PSD covariance, otherwise the eigenvalues are clipped to a threshold
        above zeros (1e-13). The default is `False` and use the clipping method as the
        Higham & Nick algorithm can be slow for large datasets.

    higham_max_iteration : int, default=100
        Maximum number of iteration of the Higham & Nick (2002) algorithm.
        The default value is `100`.

    Attributes
    ----------
    covariance_ : ndarray of shape (n_assets, n_assets)
        Estimated covariance matrix.

    Notes
    -----
    All estimators should specify all the parameters that can be set
    at the class level in their ``__init__`` as explicit keyword
    arguments (no ``*args`` or ``**kwargs``).
    """

    covariance_: np.ndarray

    @abstractmethod
    def __init__(
        self,
        nearest: bool = False,
        higham: bool = False,
        higham_max_iteration: int = 100,
    ):
        self.nearest = nearest
        self.higham = higham
        self.higham_max_iteration = higham_max_iteration

    @abstractmethod
    def fit(self, X: npt.ArrayLike, y=None):
        pass

    def _sanity_check(self, covariance: np.ndarray) -> None:
        """Perform a sanity check on the covariance matrix by verifying that all
        diagonal elements are strictly positive.
        The goal is to early detect corrupted asset data (with zero variance) that
        would lead to optimizations errors.
        """
        cond = np.diag(covariance) < 1e-15
        if np.any(cond):
            corrupted_assets = list(np.argwhere(cond).flatten())
            detail = "assets indices"
            if hasattr(self, "feature_names_in_"):
                corrupted_assets = list(self.feature_names_in_[corrupted_assets])
                detail = "assets"
            raise NonPositiveVarianceError(
                f"The following {detail} have a non positive variance:"
                f" {corrupted_assets}"
            )

    def _set_covariance(self, covariance: np.ndarray) -> None:
        """Perform checks, convert to nearest PSD if specified and saves the covariance.

        Parameters
        ----------
        covariance : array-like of shape (n_assets, n_assets)
            Estimated covariance matrix to be stored.
        """
        self._sanity_check(covariance)
        if self.nearest:
            covariance = cov_nearest(
                covariance,
                higham=self.higham,
                higham_max_iteration=self.higham_max_iteration,
            )
        # set covariance
        self.covariance_ = covariance



"""Empirical Prior Model estimator."""

# Copyright (c) 2023
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# License: BSD 3 clause

import numpy as np
import numpy.typing as npt

from skfolio.moments import BaseCovariance, BaseMu, EmpiricalCovariance, EmpiricalMu
from skfolio.prior._base import BasePrior, PriorModel
from skfolio.utils.tools import check_estimator


class EmpiricalPrior(BasePrior):
    """Empirical Prior estimator.

    The Empirical Prior estimates the :class:`~skfolio.prior.PriorModel` by fitting a
    `mu_estimator` and a `covariance_estimator` separately.

    Parameters
    ----------
    mu_estimator : BaseMu, optional
        The assets :ref:`expected returns estimator <mu_estimator>`.
        The default (`None`) is to use :class:`~skfolio.moments.EmpiricalMu`.

    covariance_estimator : BaseCovariance , optional
        The assets :ref:`covariance matrix estimator <covariance_estimator>`.
        The default (`None`) is to use  :class:`~skfolio.moments.EmpiricalCovariance`.

    is_log_normal : bool, default=False
        If this is set to True, the moments are estimated on the logarithmic returns
        as opposed to the linear returns. Then the moments estimations of the
        logarithmic returns are projected to the investment horizon and transformed
        to obtain the moments estimation of the linear returns at the investment
        horizon. If True, `investment_horizon` must be provided. The input `X` must be
        **linear returns**. They will be converted into logarithmic returns only for the
        moments estimation.

        .. seealso::

            :ref:`data preparation <data_preparation>`


    investment_horizon : float, optional
        The investment horizon used for the moments estimation of the linear returns
        when `is_log_normal` is `True`.

    Attributes
    ----------
    prior_model_ : PriorModel
        The assets :class:`~skfolio.prior.PriorModel`.

    mu_estimator_ : BaseMu
        Fitted `mu_estimator`.

    covariance_estimator_ : BaseCovariance
        Fitted `covariance_estimator`.

    n_features_in_ : int
       Number of assets seen during `fit`.

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
       Names of features seen during `fit`. Defined only when `X`
       has feature names that are all strings.

    References
    ----------
    .. [1]  "Linear vs. Compounded Returns - Common Pitfalls in Portfolio Management".
        GARP Risk Professional.
        Attilio Meucci (2010).
    """

    mu_estimator_: BaseMu
    covariance_estimator_: BaseCovariance

    def __init__(
        self,
        mu_estimator: BaseMu | None = None,
        covariance_estimator: BaseCovariance | None = None,
        is_log_normal: bool = False,
        investment_horizon: float | None = None,
    ):
        self.mu_estimator = mu_estimator
        self.covariance_estimator = covariance_estimator
        self.is_log_normal = is_log_normal
        self.investment_horizon = investment_horizon

    def fit(self, X: npt.ArrayLike, y=None) -> "EmpiricalPrior":
        """Fit the Empirical Prior estimator.

        Parameters
        ----------
        X : array-like of shape (n_observations, n_assets)
            Price returns of the assets.

        y : Ignored
            Not used, present for API consistency by convention.

        Returns
        -------
        self : EmpiricalPrior
            Fitted estimator.
        """
        self.mu_estimator_ = check_estimator(
            self.mu_estimator,
            default=EmpiricalMu(),
            check_type=BaseMu,
        )
        self.covariance_estimator_ = check_estimator(
            self.covariance_estimator,
            default=EmpiricalCovariance(),
            check_type=BaseCovariance,
        )
        # fitting estimators
        if not self.is_log_normal:
            if self.investment_horizon is not None:
                raise ValueError(
                    "`investment_horizon` must be `None` when "
                    "`is_log_normal` is `False`"
                )
            # Expected returns
            self.mu_estimator_.fit(X)
            mu = self.mu_estimator_.mu_

            # Covariance
            self.covariance_estimator_.fit(X)
            covariance = self.covariance_estimator_.covariance_
        else:
            if self.investment_horizon is None:
                raise ValueError(
                    "`investment_horizon` must be provided when "
                    "`is_log_normal` is `True`"
                )
            # Convert linear returns to log returns
            X_log = np.log(1 + X)

            # Estimates the moments on the log returns
            # Expected returns
            self.mu_estimator_.fit(X_log)
            mu = self.mu_estimator_.mu_

            # Covariance
            self.covariance_estimator_.fit(X_log)
            covariance = self.covariance_estimator_.covariance_

            # Using the property of aggregation across time we scale this distribution
            # to the investment horizon by the “square-root rule”.
            mu *= self.investment_horizon
            covariance *= self.investment_horizon

            # We convert it into a distribution of linear returns over the investment
            # horizon
            mu = np.exp(mu + 0.5 * np.diag(covariance))
            covariance = np.outer(mu, mu) * (np.exp(covariance) - 1)

        # we validate and convert to numpy after all models have been fitted to keep
        # features names information.
        X = self._validate_data(X)
        self.prior_model_ = PriorModel(
            mu=mu,
            covariance=covariance,
            returns=X,
        )
        return self



"""Base Optimization estimator."""

# Author: Hugo Delatte <delatte.hugo@gmail.com>
# License: BSD 3 clause

from abc import ABC, abstractmethod

import numpy as np
import numpy.typing as npt
import sklearn.base as skb
from sklearn.utils.validation import check_is_fitted

from skfolio.measures import RatioMeasure
from skfolio.population import Population
from skfolio.portfolio import Portfolio

# Copyright (c) 2023
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# License: BSD 3 clause
# Implementation derived from:
# scikit-portfolio, Copyright (c) 2022, Carlo Nicolini, Licensed under MIT Licence.
# scikit-learn, Copyright (c) 2007-2010 David Cournapeau, Fabian Pedregosa, Olivier
# Grisel Licensed under BSD 3 clause.


class BaseOptimization(skb.BaseEstimator, ABC):
    """Base class for all portfolio optimizations in skfolio.

    portfolio_params :  dict, optional
        Portfolio parameters passed to the portfolio evaluated by the `predict` and
        `score` methods. If not provided, the `name`, `transaction_costs`,
        `management_fees`, `previous_weights` and `risk_free_rate` are copied from the
        optimization model and passed to the portfolio.

    Attributes
    ----------
    weights_ : ndarray of shape (n_assets,) or (n_optimizations, n_assets)
        Weights of the assets.

    Notes
    -----
    All estimators should specify all the parameters that can be set
    at the class level in their `__init__` as explicit keyword
    arguments (no `*args` or `**kwargs`).
    """

    weights_: np.ndarray

    @abstractmethod
    def __init__(self, portfolio_params: dict | None = None):
        self.portfolio_params = portfolio_params

    @abstractmethod
    def fit(self, X: npt.ArrayLike, y: npt.ArrayLike | None = None):
        pass

    def predict(self, X: npt.ArrayLike) -> Portfolio | Population:
        """Predict the `Portfolio` or `Population` of `Portfolio` on `X` based on the
        fitted weights.

        Optimization estimators can return a 1D or a 2D array of `weights`.
        For a 1D array, the prediction returns a `Portfolio`.
        For a 2D array, the prediction returns a `Population` of `Portfolio`.

        If `name` is not provided in the portfolio arguments, we use the first
        500 characters of the estimator name.

        Parameters
        ----------
        X : array-like of shape (n_observations, n_assets)
            Price returns of the assets.

        Returns
        -------
        prediction : Portfolio | Population
            `Portfolio` or `Population` of `Portfolio` estimated on `X` based on the
            fitted `weights`.
        """
        check_is_fitted(self, "weights_")

        if self.portfolio_params is None:
            ptf_kwargs = {}
        else:
            ptf_kwargs = self.portfolio_params.copy()

        # Set the default portfolio parameters equal to the optimization parameters
        for param in [
            "transaction_costs",
            "management_fees",
            "previous_weights",
            "risk_free_rate",
        ]:
            if param not in ptf_kwargs and hasattr(self, param):
                ptf_kwargs[param] = getattr(self, param)

        # If 'name' is not provided in the portfolio arguments, we use the first
        # 500 characters of the optimization estimator's name
        name = ptf_kwargs.pop("name", type(self).__name__)

        # Optimization estimators can return a 1D or a 2D array of weights.
        # For a 1D array we return a portfolio.
        # For a 2D array we return a population of portfolios.
        if self.weights_.ndim == 2:
            n_portfolios = self.weights_.shape[0]
            return Population(
                [
                    Portfolio(
                        X=X,
                        weights=self.weights_[i],
                        name=f"ptf{i} - {name}",
                        **ptf_kwargs,
                    )
                    for i in range(n_portfolios)
                ]
            )
        return Portfolio(X=X, weights=self.weights_, name=name, **ptf_kwargs)

    def score(self, X: npt.ArrayLike, y: npt.ArrayLike = None) -> float:
        """Prediction score.
        If the prediction is a single `Portfolio`, the score is the Sharpe Ratio.
        If the prediction is a `Population` of `Portfolio`, the score is the mean of all
        the portfolios Sharpe Ratios in the population.

        Parameters
        ----------
        X : array-like of shape (n_observations, n_assets)
            Price returns of the assets.

        y : Ignored
            Not used, present here for API consistency by convention.

        Returns
        -------
        score : float
            The Sharpe Ratio of the portfolio if the prediction is a single `Portfolio`
            or the mean of all the portfolios Sharpe Ratios if the prediction is a
            `Population` of `Portfolio`.
        """
        result = self.predict(X)
        if isinstance(result, Population):
            return result.measures_mean(RatioMeasure.SHARPE_RATIO)
        return result.sharpe_ratio

    def fit_predict(self, X):
        """Perform `fit` on `X` and returns the predicted `Portfolio` or
        `Population` of `Portfolio` on `X` based on the fitted `weights`.
        For factor models, use `fit(X, y)` then `predict(X)` separately.

        Parameters
        ----------
        X : array-like of shape (n_observations, n_assets)
            Price returns of the assets.

        Returns
        -------
        prediction : Portfolio | Population
            `Portfolio` or `Population` of `Portfolio` estimated on `X` based on the
            fitted `weights`.
        """
        return self.fit(X).predict(X)



"""Covariance module."""

from skfolio.moments.covariance._base import (
    BaseCovariance,
)
from skfolio.moments.covariance._covariance import (
    OAS,
    DenoiseCovariance,
    DetoneCovariance,
    EWCovariance,
    EmpiricalCovariance,
    GerberCovariance,
    GraphicalLassoCV,
    LedoitWolf,
    ShrunkCovariance,
)

__all__ = [
    "BaseCovariance",
    "EmpiricalCovariance",
    "EWCovariance",
    "GerberCovariance",
    "DenoiseCovariance",
    "DetoneCovariance",
    "LedoitWolf",
    "OAS",
    "ShrunkCovariance",
    "GraphicalLassoCV",
]



"""Fast non-dominated sorting module"""

# Copyright (c) 2023
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# License: BSD 3 clause

import numpy as np

__all__ = ["dominate", "non_denominated_sort"]


def dominate(fitness_1: np.ndarray, fitness_2: np.ndarray) -> bool:
    """Compute the domination of two fitness arrays.

    Domination of `fitness_1` over `fitness_2` means that each objective (value) of
    `fitness_1` is not strictly worse than the corresponding objective of `fitness_2`
    and at least one objective is strictly better.

    Parameters
    ----------
    fitness_1 : ndarray of floats of shape (n_objectives,)
        Fitness array 1.

    fitness_2 : ndarray of floats of shape (n_objectives,)
        Fitness array 2.

    Returns
    -------
    is_dominated : bool
        Ture if `fitness_1` dominates `fitness_2`, False otherwise.
    """
    if fitness_1.ndim != fitness_2.ndim != 1:
        raise ValueError("fitness_1 and fitness_2 must be 1D array")
    not_equal = False
    for self_value, other_value in zip(fitness_1, fitness_2, strict=True):
        if self_value > other_value:
            not_equal = True
        elif self_value < other_value:
            return False
    return not_equal


def non_denominated_sort(
    fitnesses: np.ndarray, first_front_only: bool
) -> list[list[int]]:
    """Fast non-dominated sorting.

    Sort the fitnesses into different non-domination levels.
    Complexity O(MN^2) where M is the number of objectives and N the number of
    portfolios.

    Parameters
    ----------
    fitnesses: ndarray of shape(n, n_fitness)
        Fitnesses array.

    first_front_only : bool
        If this is set to True, only the first front is computed and returned.

    Returns
    -------
    fronts: list[list[int]]
      A list of Pareto fronts (lists), the first list includes non-dominated fitnesses.
    """
    n = len(fitnesses)
    fronts = []
    if n == 0:
        return fronts

    # final rank that will be returned
    n_ranked = 0
    ranked = np.array([0 for _ in range(n)])

    # for each portfolio a list of all portfolios that are dominated by this one
    is_dominating = [[x for x in range(0)] for _ in range(n)]

    # storage for the number of solutions dominated this one
    n_dominated = [0 for _ in range(n)]

    current_front = [x for x in range(0)]

    for i in range(n):
        for j in range(i + 1, n):
            if dominate(fitnesses[i], fitnesses[j]):
                is_dominating[i].append(j)
                n_dominated[j] += 1
            elif dominate(fitnesses[j], fitnesses[i]):
                is_dominating[j].append(i)
                n_dominated[i] += 1

        if n_dominated[i] == 0:
            current_front.append(i)
            ranked[i] = 1.0
            n_ranked += 1

    # append the first front to the current front
    fronts.append(current_front)

    if first_front_only:
        return fronts

    # while not all solutions are assigned to a pareto front
    while n_ranked < n:
        next_front = []
        # for each portfolio in the current front
        for i in current_front:
            # all solutions that are dominated by this portfolio
            for j in is_dominating[i]:
                n_dominated[j] -= 1
                if n_dominated[j] == 0:
                    next_front.append(j)
                    ranked[j] = 1.0
                    n_ranked += 1

        fronts.append(next_front)
        current_front = next_front

    return fronts



"""Portfolio module.
`Portfolio` and `MultiPeriodPortfolio` objects are returned by the `predict` method of
Optimization estimators.
They need to be homogenous to the convex optimization problems meaning that `Portfolio`
is the dot product of the assets weights with the assets returns and
`MultiPeriodPortfolio` is a list of `Portfolio`.
"""

from skfolio.portfolio._base import BasePortfolio
from skfolio.portfolio._multi_period_portfolio import MultiPeriodPortfolio
from skfolio.portfolio._portfolio import Portfolio

__all__ = ["BasePortfolio", "Portfolio", "MultiPeriodPortfolio"]
